from typing import Literal
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple  # Classes
from typing import Union

import anndata
import numpy as np
import pandas as pd
import scanpy as sc
from anndata import AnnData
from matplotlib import axes
from matplotlib import gridspec
from matplotlib import pyplot as pl
from matplotlib.axes import Axes
from matplotlib.figure import Figure  # SubplotParams as sppars,

_VarNames = Union[str, Sequence[str]]
ColorLike = Union[str, Tuple[float, ...]]


class _AxesSubplot(Axes, axes.SubplotBase):
    """Intersection between Axes and SubplotBase: Has methods of both"""


# -------------------------------------------------------------------------------
# Simple plotting functions
# -------------------------------------------------------------------------------
# We build the MaskedDotPlot based on sc.pl.DotPlot. Compared to scanpy's dotplot
# the only add-ons are identifiers around dots e.g. for tree genes and marker genes.
# Unfortunately a lot of class methods needed to be copied and sligthly modified.
# Maybe this could be done more efficiently. Also the legend add-ons for the identifiers
# have some positioning problems which need to be fixed.


# def matrix(
#    matrix,
#    xlabel=None,
#    ylabel=None,
#    xticks=None,
#    yticks=None,
#    title=None,
#    colorbar_shrink=0.5,
#    color_map=None,
#    show=None,
#    save=None,
#    ax=None,
# ):
#    """Plot a matrix."""
#    if ax is None:
#        ax = pl.gca()
#    img = ax.imshow(matrix, cmap=color_map)
#    if xlabel is not None:
#        ax.set_xlabel(xlabel)
#    if ylabel is not None:
#        ax.set_ylabel(ylabel)
#    if title is not None:
#        ax.set_title(title)
#    if xticks is not None:
#        ax.set_xticks(range(len(xticks)), xticks, rotation="vertical")
#    if yticks is not None:
#        ax.set_yticks(range(len(yticks)), yticks)
#    pl.colorbar(img, shrink=colorbar_shrink, ax=ax)  # need a figure instance for colorbar
#    savefig_or_show("matrix", show=show, save=save)


def fix_kwds(kwds_dict: dict, **kwargs):
    """
    Given a dictionary of plot parameters (kwds_dict) and a dict of kwds,
    merge the parameters into a single consolidated dictionary to avoid
    argument duplication errors.
    If kwds_dict an kwargs have the same key, only the value in kwds_dict is kept.

    Args:
        kwds_dict:
            kwds_dictionary
        kwargs

    Returns:
        kwds_dict merged with kwargs

    Examples::

        >>> def _example(**kwds):
             return fix_kwds(kwds, key1="value1", key2="value2")
        >>> example(key1="value10", key3="value3")
            {'key1': 'value10, 'key2': 'value2', 'key3': 'value3'}

    """

    kwargs.update(kwds_dict)

    return kwargs


def make_grid_spec(
    ax_or_figsize: Union[Tuple[int, int], _AxesSubplot],
    nrows: int,
    ncols: int,
    wspace: Optional[float] = None,
    hspace: Optional[float] = None,
    width_ratios: Optional[Sequence[float]] = None,
    height_ratios: Optional[Sequence[float]] = None,
) -> Tuple[Figure, gridspec.GridSpecBase]:
    kw = dict(
        wspace=wspace,
        hspace=hspace,
        width_ratios=width_ratios,
        height_ratios=height_ratios,
    )
    if isinstance(ax_or_figsize, tuple):
        fig = pl.figure(figsize=ax_or_figsize)
        return fig, gridspec.GridSpec(nrows, ncols, **kw)
    else:
        ax = ax_or_figsize
        ax.axis("off")
        ax.set_frame_on(False)
        ax.set_xticks([])
        ax.set_yticks([])
        return ax.figure, ax.get_subplotspec().subgridspec(nrows, ncols, **kw)


def _get_basis(adata: anndata.AnnData, basis: str):

    if basis in adata.obsm.keys():
        basis_key = basis

    elif f"X_{basis}" in adata.obsm.keys():
        basis_key = f"X_{basis}"

    return basis_key


class MaskedDotPlot(sc.pl.DotPlot):
    """Dotplot with additional annotation masks

    TODO: Add full description here
    """

    def __init__(
        self,
        adata: AnnData,
        var_names: Union[_VarNames, Mapping[str, _VarNames]],
        groupby: Union[str, Sequence[str]],
        tree_genes: Optional[dict] = None,
        marker_genes: Optional[dict] = None,
        further_celltypes: Optional[Sequence[str]] = None,
        tree_genes_color: Optional[str] = "darkblue",
        marker_genes_color: Optional[str] = "green",
        non_adata_celltypes_color: Optional[str] = "grey",
        use_raw: Optional[bool] = None,
        log: bool = False,
        num_categories: int = 7,
        categories_order: Optional[Sequence[str]] = None,
        title: Optional[str] = None,
        figsize: Optional[Tuple[float, float]] = None,
        gene_symbols: Optional[str] = None,
        var_group_positions: Optional[Sequence[Tuple[int, int]]] = None,
        var_group_labels: Optional[Sequence[str]] = None,
        var_group_rotation: Optional[float] = None,
        layer: Optional[str] = None,
        expression_cutoff: float = 0.0,
        mean_only_expressed: bool = False,
        standard_scale: Literal["var", "group"] = "var",
        dot_color_df: Optional[pd.DataFrame] = None,
        dot_size_df: Optional[pd.DataFrame] = None,
        ax: Optional[_AxesSubplot] = None,
        grid: Optional[bool] = True,
        grid_linewidth: Optional[float] = 0.1,
        **kwds,
    ):
        self.tree_genes = tree_genes
        self.marker_genes = marker_genes
        self.further_celltypes = further_celltypes
        self.show_marker_legend = True
        self.tree_genes_color = tree_genes_color
        self.marker_genes_color = marker_genes_color
        self.non_adata_celltypes_color = non_adata_celltypes_color
        self.grid_linewidth = grid_linewidth
        sc.pl.DotPlot.__init__(
            self,
            adata,
            var_names,
            groupby,
            use_raw=use_raw,
            log=log,
            num_categories=num_categories,
            categories_order=categories_order,
            title=title,
            figsize=figsize,
            gene_symbols=gene_symbols,
            var_group_positions=var_group_positions,
            var_group_labels=var_group_labels,
            var_group_rotation=var_group_rotation,
            layer=layer,
            ax=ax,
            expression_cutoff=expression_cutoff,
            mean_only_expressed=mean_only_expressed,
            standard_scale=standard_scale,
            dot_color_df=dot_color_df,
            dot_size_df=dot_size_df,
            **kwds,
        )
        self.style(
            cmap="Reds",
            color_on="on_dot",
            # 'square',
            # dot_max=None,
            # dot_min=None,
            # smallest_dot=0.0,
            # largest_dot=200.0,
            # dot_edge_color='black',
            # dot_edge_lw=0.2,
            # size_exponent=1.5,
            grid=grid,  # True,
            # x_padding=0.8,
            # y_padding=1.0)
        )
        # default height and width are 0.35 and 0.37. Since we use symmetrical markers we set them to the same values
        # (this restricts the flexibility of the function... we'd like to keep the generality of sc.pl.dotplot, maybe
        # this can be done better)
        default_length = (self.DEFAULT_CATEGORY_WIDTH + self.DEFAULT_CATEGORY_HEIGHT) / 2  # type: ignore
        self.DEFAULT_CATEGORY_WIDTH = default_length
        self.DEFAULT_CATEGORY_HEIGHT = default_length

        if further_celltypes:
            self.DEFAULT_CATEGORY_HEIGHT = self.DEFAULT_CATEGORY_HEIGHT * (
                (len(further_celltypes) + len(self.categories)) / (len(self.categories))
            )

    def _mainplot(self, ax):
        # work on a copy of the dataframes. This is to avoid changes
        # on the original data frames after repetitive calls to the
        # DotPlot object, for example once with swap_axes and other without

        _color_df = self.dot_color_df.copy()
        _size_df = self.dot_size_df.copy()

        if self.var_names_idx_order is not None:
            _color_df = _color_df.iloc[:, self.var_names_idx_order]
            _size_df = _size_df.iloc[:, self.var_names_idx_order]

        if self.categories_order is not None:
            _color_df = _color_df.loc[self.categories_order, :]
            _size_df = _size_df.loc[self.categories_order, :]

        if self.further_celltypes:
            _other_celltypes_df = pd.DataFrame(index=_color_df.index, columns=_color_df.columns, data=False)
            for ct in self.further_celltypes:
                _color_df.loc[ct] = 0
                _size_df.loc[ct] = 0
                _other_celltypes_df.loc[ct] = True
        else:
            _other_celltypes_df = None
        if self.tree_genes:
            _trees_df = pd.DataFrame(index=_color_df.index, columns=_color_df.columns, data=False)
            for ct, genes in self.tree_genes.items():
                _trees_df.loc[ct, genes] = True
        else:
            _trees_df = None
        if self.marker_genes:
            _marker_df = pd.DataFrame(index=_color_df.index, columns=_color_df.columns, data=False)
            for ct, genes in self.marker_genes.items():
                _marker_df.loc[ct, genes] = True
        else:
            _marker_df = None

        if self.are_axes_swapped:
            _size_df = _size_df.T
            _color_df = _color_df.T
            if _other_celltypes_df:
                _other_celltypes_df = _other_celltypes_df.T
            if _trees_df:
                _trees_df = _trees_df.T
            if _marker_df:
                _marker_df = _marker_df.T
        self.cmap = self.kwds.get("cmap", self.cmap)
        if "cmap" in self.kwds:
            del self.kwds["cmap"]

        normalize, dot_min, dot_max = self._dotplot(
            _size_df,
            _color_df,
            ax,
            _trees_df,
            _marker_df,
            _other_celltypes_df,
            tree_genes_color=self.tree_genes_color,
            marker_genes_color=self.marker_genes_color,
            non_adata_celltypes_color=self.non_adata_celltypes_color,
            cmap=self.cmap,
            dot_max=self.dot_max,
            dot_min=self.dot_min,
            color_on=self.color_on,
            edge_color=self.dot_edge_color,
            edge_lw=self.dot_edge_lw,
            smallest_dot=self.smallest_dot,
            largest_dot=self.largest_dot,
            size_exponent=self.size_exponent,
            grid=self.grid,
            grid_linewidth=self.grid_linewidth,
            x_padding=self.plot_x_padding,
            y_padding=self.plot_y_padding,
            **self.kwds,
        )

        self.dot_min, self.dot_max = dot_min, dot_max
        return normalize

    @staticmethod
    def _dotplot(
        dot_size,
        dot_color,
        dot_ax,
        tree_genes,
        marker_genes,
        grey_celltypes,
        tree_genes_color="darkblue",
        marker_genes_color="green",
        non_adata_celltypes_color="grey",
        cmap: str = "Reds",
        color_on: Optional[str] = "dot",
        y_label: Optional[str] = None,
        dot_max: Optional[float] = None,
        dot_min: Optional[float] = None,
        standard_scale: Literal["var", "group"] = None,
        smallest_dot: Optional[float] = 0.0,
        largest_dot: Optional[float] = 200,
        size_exponent: Optional[float] = 2,
        edge_color: Optional[ColorLike] = None,
        edge_lw: Optional[float] = None,
        grid: Optional[bool] = False,
        grid_linewidth: Optional[float] = 0.1,
        x_padding: Optional[float] = 0.8,
        y_padding: Optional[float] = 1.0,
        **kwds,
    ):
        """\
        Makes a *dot plot* given two data frames, one containing
        the doc size and other containing the dot color. The indices and
        columns of the data frame are used to label the output image
        The dots are plotted using :func:`matplotlib.pyplot.scatter`. Thus, additional
        arguments can be passed.
        Args:
            dot_size: Data frame containing the dot_size.
            dot_color: Data frame containing the dot_color, should have the same,
                    shape, columns and indices as dot_size.
            dot_ax: matplotlib axis
            tree_genes_color
                Color for tree genes.
            marker_genes_color
                Color for marker genes
            non_adata_celltypes_color
                Color for celltypes that don't occur in adata
            cmap
                String denoting matplotlib color map.
            color_on
                Options are 'dot' or 'square'. Be default the colomap is applied to
                the color of the dot. Optionally, the colormap can be applied to an
                square behind the dot, in which case the dot is transparent and only
                the edge is shown.
            y_label: String. Label for y axis
            dot_max
                If none, the maximum dot size is set to the maximum fraction value found
                (e.g. 0.6). If given, the value should be a number between 0 and 1.
                All fractions larger than dot_max are clipped to this value.
            dot_min
                If none, the minimum dot size is set to 0. If given,
                the value should be a number between 0 and 1.
                All fractions smaller than dot_min are clipped to this value.
            standard_scale
                Whether or not to standardize that dimension between 0 and 1,
                meaning for each variable or group,
                subtract the minimum and divide each by its maximum.
            smallest_dot
                If none, the smallest dot has size 0.
                All expression levels with `dot_min` are plotted with this size.
            edge_color
                Dot edge color. When `color_on='dot'` the default is no edge. When
                `color_on='square'`, edge color is white
            edge_lw
                Dot edge line width. When `color_on='dot'` the default is no edge. When
                `color_on='square'`, line width = 1.5
            grid
                Adds a grid to the plot
            x_paddding
                Space between the plot left/right borders and the dots center. A unit
                is the distance between the x ticks. Only applied when color_on = dot
            y_paddding
                Space between the plot top/bottom borders and the dots center. A unit is
                the distance between the y ticks. Only applied when color_on = dot
            kwds
                Are passed to :func:`matplotlib.pyplot.scatter`.
        Returns:
            matplotlib.colors.Normalize, dot_min, dot_max
        """
        assert dot_size.shape == dot_color.shape, (
            "please check that dot_size " "and dot_color dataframes have the same shape"
        )

        assert list(dot_size.index) == list(dot_color.index), (
            "please check that dot_size " "and dot_color dataframes have the same index"
        )

        assert list(dot_size.columns) == list(dot_color.columns), (
            "please check that the dot_size " "and dot_color dataframes have the same columns"
        )

        if standard_scale == "group":
            dot_color = dot_color.sub(dot_color.min(1), axis=0)
            dot_color = dot_color.div(dot_color.max(1), axis=0).fillna(0)
        elif standard_scale == "var":
            dot_color -= dot_color.min(0)
            dot_color = (dot_color / dot_color.max(0)).fillna(0)
        elif standard_scale is None:
            pass

        # make scatter plot in which
        # x = var_names
        # y = groupby category
        # size = fraction
        # color = mean expression

        # +0.5 in y and x to set the dot center at 0.5 multiples
        # this facilitates dendrogram and totals alignment for
        # matrixplot, dotplot and stackec_violin using the same coordinates.
        y, x = np.indices(dot_color.shape)
        y = y.flatten() + 0.5
        x = x.flatten() + 0.5
        frac = dot_size.values.flatten()
        mean_flat = dot_color.values.flatten()
        if isinstance(grey_celltypes, pd.DataFrame):
            grey_celltypes_mask = grey_celltypes.values.flatten()
        if isinstance(tree_genes, pd.DataFrame):
            tree_genes_mask = tree_genes.values.flatten()
        if isinstance(marker_genes, pd.DataFrame):
            marker_genes_mask = marker_genes.values.flatten()
        cmap = pl.get_cmap(kwds.get("cmap", cmap))
        if "cmap" in kwds:
            del kwds["cmap"]
        if dot_max is None:
            dot_max = np.ceil(max(frac) * 10) / 10
        else:
            if dot_max < 0 or dot_max > 1:
                raise ValueError("`dot_max` value has to be between 0 and 1")
        if dot_min is None:
            dot_min = 0
        else:
            if dot_min < 0 or dot_min > 1:
                raise ValueError("`dot_min` value has to be between 0 and 1")

        if dot_min != 0 or dot_max != 1:
            # clip frac between dot_min and  dot_max
            frac = np.clip(frac, dot_min, dot_max)
            old_range = dot_max - dot_min
            # re-scale frac between 0 and 1
            frac = (frac - dot_min) / old_range

        size = frac ** size_exponent
        # rescale size to match smallest_dot and largest_dot
        size = size * (largest_dot - smallest_dot) + smallest_dot  # type: ignore

        import matplotlib.colors

        normalize = matplotlib.colors.Normalize(vmin=kwds.get("vmin"), vmax=kwds.get("vmax"))

        if color_on == "square":
            if edge_color is None:
                from seaborn.utils import relative_luminance

                # use either black or white for the edge color
                # depending on the luminance of the background
                # square color
                edge_color = []  # type: ignore
                for color_value in cmap(normalize(mean_flat)):  # type: ignore
                    lum = relative_luminance(color_value)
                    edge_color.append(".15" if lum > 0.408 else "w")  # type: ignore

            edge_lw = 1.5 if edge_lw is None else edge_lw

            # first make a heatmap similar to `sc.pl.matrixplot`
            # (squares with the asigned colormap). Circles will be plotted
            # on top
            dot_ax.pcolor(dot_color.values, cmap=cmap, norm=normalize)
            for axis in ["top", "bottom", "left", "right"]:
                dot_ax.spines[axis].set_linewidth(1.5)
            kwds = fix_kwds(
                kwds,
                s=size,
                cmap=cmap,
                norm=None,
                linewidth=edge_lw,
                facecolor="none",
                edgecolor=edge_color,
            )
            dot_ax.scatter(x, y, **kwds)
        else:
            edge_color = "none" if edge_color is None else edge_color
            edge_lw = 0.0 if edge_lw is None else edge_lw

            # TODO: i think the factor 1.5 can be infered from some variable, but i don't know which...
            # In case you want to make the edges nicer consider also plotting white overlapping squares on all genes
            # that are not tree genes so that the outer edge half is not plotted (white square is overlapping). The
            # correct order of the (then) four scatter plots is important
            if isinstance(tree_genes, pd.DataFrame):
                dot_ax.scatter(
                    x[tree_genes_mask],
                    y[tree_genes_mask],
                    s=np.max(size) * 1.55,
                    marker="s",
                    color="none",
                    linewidth=2,
                    edgecolor=tree_genes_color,
                )
                # dot_ax.scatter(x[~tree_genes_mask], y[~tree_genes_mask], s=np.max(size)*1.85, marker='s',
                # color='white',edgecolor='none')
                # Tried it, but without the second line and 1.55 instead of 1.85 just looks better
            if isinstance(grey_celltypes, pd.DataFrame):
                dot_ax.scatter(
                    x[grey_celltypes_mask],
                    y[grey_celltypes_mask],
                    s=np.max(size) * 1.95,
                    marker="s",
                    color=non_adata_celltypes_color,
                    edgecolor="none",
                    alpha=0.3,
                )
            if isinstance(marker_genes, pd.DataFrame):
                dot_ax.scatter(
                    x[marker_genes_mask],
                    y[marker_genes_mask],
                    s=np.max(size) * 1.85,
                    marker="s",
                    color=marker_genes_color,
                    edgecolor="none",
                    alpha=0.3,
                )

            color = cmap(normalize(mean_flat))  # type: ignore
            kwds = fix_kwds(
                kwds,
                s=size,
                cmap=cmap,
                color=color,
                norm=None,
                linewidth=edge_lw,
                edgecolor=edge_color,
            )

            dot_ax.scatter(x, y, **kwds)

        y_ticks = np.arange(dot_color.shape[0]) + 0.5
        dot_ax.set_yticks(y_ticks)
        dot_ax.set_yticklabels([dot_color.index[idx] for idx, _ in enumerate(y_ticks)], minor=False)

        x_ticks = np.arange(dot_color.shape[1]) + 0.5
        dot_ax.set_xticks(x_ticks)
        dot_ax.set_xticklabels(
            [dot_color.columns[idx] for idx, _ in enumerate(x_ticks)],
            rotation=90,
            ha="center",
            minor=False,
        )
        dot_ax.tick_params(axis="both", labelsize="small")
        dot_ax.grid(False)
        dot_ax.set_ylabel(y_label)

        # to be consistent with the heatmap plot, is better to
        # invert the order of the y-axis, such that the first group is on
        # top
        dot_ax.set_ylim(dot_color.shape[0], 0)
        dot_ax.set_xlim(0, dot_color.shape[1])

        if color_on == "dot":
            # add padding to the x and y lims when the color is not in the square
            # default y range goes from 0.5 to num cols + 0.5
            # and default x range goes from 0.5 to num rows + 0.5, thus
            # the padding needs to be corrected.
            x_padding = x_padding - 0.5  # type: ignore
            y_padding = y_padding - 0.5  # type: ignore
            dot_ax.set_ylim(dot_color.shape[0] + y_padding, -y_padding)

            dot_ax.set_xlim(-x_padding, dot_color.shape[1] + x_padding)

        if grid:
            dot_ax.grid(True, color="gray", linewidth=grid_linewidth)  # linewidth=0.1)
            dot_ax.set_axisbelow(True)

        return normalize, dot_min, dot_max

    def _plot_marker_legend(
        self,
        marker_legend_ax: Axes,
        tree_genes_color="darkblue",
        marker_genes_color="green",
        non_adata_celltypes_color="grey",
        i=0,
    ):
        """New"""
        size = self.dot_max * 1.55 * 150

        ax = marker_legend_ax
        if i == 0:
            ax.scatter([0], [0], s=size, marker="s", color="none", linewidth=2, edgecolor=tree_genes_color)
            ax.set_title("Spapros marker", size="small")
        elif i == 1:
            ax.scatter([0], [0], s=size, marker="s", color=marker_genes_color, edgecolor="none", alpha=0.3)
            ax.set_title("DE or lit. gene", size="small")
        elif i == 2:
            ax.scatter([0], [0], s=size, marker="s", color=non_adata_celltypes_color, edgecolor="none", alpha=0.3)
            ax.set_title("celltype not in dataset", size="small")

        # ymax = ax.get_ylim()[1]
        ax.set_ylim(-10, 10)
        # size_legend_ax.set_title(self.size_title, y=ymax + 0.45, size='small')

        # ax.set_ylim([-1.5,1.5])
        # ax.set_yticks([1,0,-1])
        ax.set_xticks([])
        ax.set_yticks([])
        # ax.yaxis.set_label_position("right")
        # ax.tick_params(labelsize='small',length=0) # title fontsize
        # ax.yaxis.tick_right()
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.grid(False)

    def _plot_legend(self, legend_ax, return_ax_dict, normalize):

        # to maintain the fixed height size of the legends, a
        # spacer of variable height is added at the bottom.
        # The structure for the legends is:
        # first row: variable space to keep the other rows of
        #            the same size (avoid stretching)
        # second row: legend for dot size
        # third row: spacer to avoid color and size legend titles to overlap
        # fourth row: colorbar

        cbar_legend_height = self.min_figure_height * 0.08
        size_legend_height = self.min_figure_height * 0.27
        marker_legend_height = self.min_figure_height * 0.17  # 0.20
        spacer_height = self.min_figure_height * 0.3
        small_spacer_height = self.min_figure_height * 0.12  # 0.15

        height_ratios = [
            # self.height - size_legend_height - cbar_legend_height - spacer_height,
            self.height
            - 3 * marker_legend_height
            - size_legend_height
            - cbar_legend_height
            - spacer_height  # - 2 * spacer_height
            - 3 * small_spacer_height,  # - 2 * small_spacer_height,
            marker_legend_height,
            small_spacer_height,
            marker_legend_height,
            small_spacer_height,
            marker_legend_height,
            small_spacer_height,  # spacer_height,
            size_legend_height,
            spacer_height,
            cbar_legend_height,
        ]
        fig, legend_gs = make_grid_spec(legend_ax, nrows=10, ncols=1, height_ratios=height_ratios)

        ### Added
        if self.show_marker_legend:
            for i in range(3):
                marker_legend_ax = fig.add_subplot(legend_gs[i * 2 + 1])  # legend_gs[i + 1 + int(i > 0)])
                self._plot_marker_legend(
                    marker_legend_ax,
                    tree_genes_color=self.tree_genes_color,
                    marker_genes_color=self.marker_genes_color,
                    non_adata_celltypes_color=self.non_adata_celltypes_color,
                    i=i,
                )
                return_ax_dict[f"marker_legend_ax_{i}"] = marker_legend_ax

        if self.show_size_legend:
            size_legend_ax = fig.add_subplot(legend_gs[7])
            self._plot_size_legend(size_legend_ax)
            return_ax_dict["size_legend_ax"] = size_legend_ax

        if self.show_colorbar:
            color_legend_ax = fig.add_subplot(legend_gs[9])
            self._plot_colorbar(color_legend_ax, normalize)
            return_ax_dict["color_legend_ax"] = color_legend_ax



"""Plotting Module."""
import itertools
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union

import matplotlib
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import scipy.cluster.hierarchy as sch
import seaborn as sns
from matplotlib.gridspec import GridSpec
from scipy.interpolate import interp1d
from spapros.plotting._masked_dotplot import MaskedDotPlot
from upsetplot import from_indicators
from upsetplot import UpSet
from venndata import venn


#############################
## evaluation related plots ##
#############################


def ordered_confusion_matrices(conf_mats: List[pd.DataFrame]) -> List[pd.DataFrame]:
    """Rearranges confusion matrices by a linkage clustering.

    Notes:
        The matrices in conf_mats must have the same indices (and columns). We calculate the clustering on
        a concatenated list of confusion matrices and then reorder each matrix by the resulting order.

    Args:
        conf_mats:
            The confusion matrices to be reordered.

    Returns:
        list of pd.DataFrame:
            Reordered confusion matrices.
    """

    # TODO maybe move to utils ?

    pooled = pd.concat(conf_mats, axis=1)

    pairwise_distances = sch.distance.pdist(pooled)
    linkage = sch.linkage(pairwise_distances, method="complete")
    cluster_distance_threshold = pairwise_distances.max() / 2
    idx_to_cluster_array = sch.fcluster(linkage, cluster_distance_threshold, criterion="distance")
    idx = np.argsort(idx_to_cluster_array)

    ordered_mats = [conf_mat.copy().iloc[idx, :].T.iloc[idx, :].T for conf_mat in conf_mats]
    return ordered_mats


def confusion_matrix(
    set_ids: List[str],
    conf_matrices: Dict[str, pd.DataFrame],
    ordered: Union[bool, list] = True,
    show: bool = True,
    save: Union[bool, str] = False,
    size_factor: float = 6,
    fontsize: int = 18,
    n_cols: int = 3,
    rotate_x_labels: bool = True,
) -> None:
    """Plot heatmap of cell type classification confusion matrices.

    Args:
        set_ids:
            List of probe set ids.
        conf_matrices:
            Confusion matrix of each probe set given in ``set_ids``.
        ordered: bool or list
            If set to True a linkage clustering is computed to order cell types together that are hard to distinguish.
            If multiple set_ids are provided the same order is applied to all of them.
            Alternatively provide a list with a custom order.
        show:
            Show the figure.
        save:
            If `True` or a `str`, save the figure.
        size_factor:
             Factor for scaling the figure size.
        fontsize:
            Matplotlib fontsize.
        n_cols:
            Number of subplot columns.
        rotate_x_labels:
            Rotate the xticklabels by 45 degrees.
    """

    n_plots = len(set_ids)
    n_rows = (n_plots // n_cols) + int((n_plots % n_cols) > 0)

    if ordered:
        if isinstance(ordered, bool):
            tmp_matrices = [conf_matrices[set_id].copy() for set_id in set_ids]
            cms = ordered_confusion_matrices(tmp_matrices)
        elif isinstance(ordered, list):
            cms = [conf_matrices[set_id].copy().loc[ordered, ordered] for set_id in set_ids]
    else:
        cms = [conf_matrices[set_id].copy() for set_id in set_ids]

    fig = plt.figure(figsize=(size_factor * n_cols, 0.75 * size_factor * n_rows))
    for i, set_id in enumerate(set_ids):
        ax = plt.subplot(n_rows, n_cols, i + 1)
        sns.heatmap(
            cms[i], cmap="OrRd", cbar=False, ax=ax, vmin=0, vmax=1, annot=True, fmt=".2f", annot_kws={"size": fontsize}
        )
        # sns.heatmap(cms[i],cmap="OrRd",cbar=(i == (len(set_ids)-1)),ax=ax,vmin=0,vmax=1,annot=True,fmt=".2f")
        if i == 0:
            plt.tick_params(axis="both", which="major", bottom=False, labelbottom=False, top=True, labeltop=True)
            if rotate_x_labels:
                plt.setp(plt.gca().get_xticklabels(), ha="left", rotation=45)
        elif (i % n_cols) == 0:
            plt.tick_params(
                axis="both",
                which="major",
                bottom=False,
                labelbottom=False,
                top=False,
                labeltop=False,
                left=True,
                labelleft=True,
            )
        elif (i // n_cols) < 1:
            plt.tick_params(
                axis="both",
                which="major",
                bottom=False,
                labelbottom=False,
                top=True,
                labeltop=True,
                left=False,
                labelleft=False,
            )
            if rotate_x_labels:
                plt.setp(plt.gca().get_xticklabels(), ha="left", rotation=45)
        else:
            plt.tick_params(
                axis="both",
                which="major",
                bottom=False,
                labelbottom=False,
                top=False,
                labeltop=False,
                left=False,
                labelleft=False,
            )

        plt.title(set_id)
        ax.title.set_fontsize(fontsize)
        ax.tick_params(axis="both", labelsize=fontsize)

    # plt.subplots_adjust(top=1.54, bottom=0.08, left=0.05, right=0.95, hspace=0.20, wspace=0.25)
    plt.tight_layout()
    if show:
        plt.show()
    if save:
        fig.savefig(save, bbox_inches="tight", transparent=True)
    plt.close()


def correlation_matrix(
    set_ids: List[str],
    cor_matrices: Dict[str, pd.DataFrame],
    show: bool = True,
    save: Union[bool, str] = False,
    size_factor: float = 5,
    fontsize: int = 28,
    n_cols: int = 3,
    colorbar: bool = True,
) -> None:
    """Plot heatmap of gene correlation matrix.

    Args:
        set_ids:
            List of probe set ids.
        cor_matrices:
            Ordered correlation matrix of each probe set given in ``set_ids``.
        show:
            Show the figure.
        save:
            Save plot to path.
        size_factor:
            Factor for scaling the figure size.
        fontsize:
            Matplotlib fontsize.
        colorbar:
            Whether to draw a colorbar.
        n_cols:
            Number of subplot columns.
    """

    n_plots = len(set_ids)
    n_rows = (n_plots // n_cols) + int((n_plots % n_cols) > 0)

    HSPACE_INCHES = 1 * n_rows
    WSPACE_INCHES = 0.5 * n_cols
    TOP_INCHES = 1
    BOTTOM_INCHES = 0.5
    RIGHT_INCHES = 2
    LEFT_INCHES = 0.5
    CBAR_WIDTH_INCHES = 0.4
    CBAR_RIGHT_INCHES = 1
    CBAR_LEFT_INCHES = RIGHT_INCHES - CBAR_WIDTH_INCHES - CBAR_RIGHT_INCHES

    FIGURE_WIDTH = (size_factor * n_cols) + (((n_cols - 1) / n_cols) * WSPACE_INCHES) + RIGHT_INCHES + LEFT_INCHES
    FIGURE_HEIGHT = (size_factor * n_rows) + (((n_rows - 1) / n_rows) * HSPACE_INCHES) + TOP_INCHES + BOTTOM_INCHES

    fig = plt.figure(figsize=(FIGURE_WIDTH, FIGURE_HEIGHT))

    for i, set_id in enumerate(set_ids):
        plt.subplot(n_rows, n_cols, i + 1)
        plt.imshow(cor_matrices[set_id].values, cmap="seismic", vmin=-1, vmax=1)
        plt.title(set_id, fontsize=fontsize)
        ax = plt.gca()
        ax.axes.get_xaxis().set_visible(False)
        ax.axes.get_yaxis().set_visible(False)

    HSPACE = HSPACE_INCHES / FIGURE_HEIGHT
    WSPACE = WSPACE_INCHES / FIGURE_WIDTH
    TOP = 1 - (TOP_INCHES / FIGURE_HEIGHT)
    BOTTOM = BOTTOM_INCHES / FIGURE_HEIGHT
    RIGHT = 1 - (RIGHT_INCHES / FIGURE_WIDTH)
    LEFT = LEFT_INCHES / FIGURE_WIDTH
    SUBPLOT_HEIGHT = size_factor / FIGURE_HEIGHT
    CBAR_WIDTH = CBAR_WIDTH_INCHES / FIGURE_WIDTH
    CBAR_RIGHT = 1 - (CBAR_RIGHT_INCHES / FIGURE_WIDTH)
    CBAR_LEFT = CBAR_RIGHT - (CBAR_LEFT_INCHES / FIGURE_WIDTH)

    plt.subplots_adjust(bottom=BOTTOM, top=TOP, left=LEFT, right=RIGHT, hspace=HSPACE, wspace=WSPACE)

    if colorbar:
        cbar_ax = fig.add_axes([CBAR_LEFT, TOP - SUBPLOT_HEIGHT, CBAR_WIDTH, SUBPLOT_HEIGHT])
        cbar = plt.colorbar(cax=cbar_ax)
        cbar.ax.tick_params(labelsize=fontsize)

    if show:
        plt.show()
    if save:
        fig.savefig(save, bbox_inches="tight", transparent=True)
    plt.close()


def marker_correlation(
    marker_corr: Dict[str, pd.DataFrame],
    corr_metric: str = "per marker mean > 0.025",
    mask_below_max: bool = False,
    set_ids: Optional[List[str]] = None,
    rename_set_ids: Dict[str, str] = {},
    size_factor: float = 1.0,
    save: Optional[str] = None,
) -> None:
    """Plot maximal correlations with marker genes

    Args:
        marker_corr:
            For each set id a Dataframe containing maximal correlations for each marker gene. The index contains the marker
            genes symbols and the Dataframe needs to have one column named ``corr_metric`` with float values.
        corr_metric:
            Column in dataframes of ``marker_corr`` that contains the correlation values.
        mask_below_max:
            Whether to mask values that are lower than the maximum correlation with a marker gene for each cell type.
            The masking applies per set id.
        set_ids:
            Plot subset of set ids (must occur as keys in ``marker_corr``).
        rename_set_ids:
            Mapping to rename set_ids for tick labels.
        size_factor:
            Scale figure size.
        save:
            Optionally save figure to path.

    """

    #######################
    # Prepare data matrix #
    #######################

    if set_ids is None:
        set_ids = list(marker_corr.keys())
    set_id_names = [rename_set_ids[s] if (s in rename_set_ids) else s for s in set_ids]

    # Sort genes alphabetically by cell type group and within group alphabetically by gene name
    df_set0 = marker_corr[set_ids[0]]
    df_set0["genes"] = df_set0.index
    df_set0 = df_set0.sort_values(["celltype", "genes"])
    df_set0 = df_set0.loc[~df_set0["per marker mean > 0.025"].isnull()]
    genes = df_set0["genes"].values

    # Create dataframe for heatmap
    df = pd.DataFrame(
        index=set_id_names,
        data={gene: [marker_corr[set_id].loc[gene, corr_metric] for set_id in set_ids] for gene in genes},
    )

    # Create dataframe for masked heatmap
    df_masked = df.copy()
    for ct in df_set0["celltype"].unique():
        genes_tmp = df_set0.loc[df_set0["celltype"] == ct].index.tolist()
        for method, max_val in df[genes_tmp].max(axis=1).items():
            genes_reduced = [g for g in genes_tmp if (df.loc[method, g] != max_val)]
            df_masked.loc[method, genes_reduced] = np.nan

    ########
    # Plot #
    ########

    # plot parameters
    width_per_gene = 0.25
    height_per_set = 0.3
    pos_factor = 0.02

    # get positions for group brackets
    group_positions = []
    group_labels = []
    for ct in df_set0["celltype"].unique():
        tmp = pd.Series((df_set0["celltype"] == ct).values)
        group_positions.append([tmp.idxmax(), tmp.where(tmp).last_valid_index()])
        group_labels.append(ct)

    # Set nrows, ncols and figsize
    ncols_genes = len(df.columns)
    ncols_cbar = 2  # max(int(0.03 * ncols), 1)
    ncols = ncols_genes + ncols_cbar
    nrows = len(df) + 1  # 1 for group brackets
    fig = plt.figure(figsize=((width_per_gene * ncols) * size_factor, (height_per_set * nrows) * size_factor))

    # Define axes
    ax1 = plt.subplot2grid((nrows, ncols), (1, 0), colspan=ncols_genes, rowspan=nrows - 1)
    ax2 = plt.subplot2grid((nrows, ncols), (1, ncols_genes), colspan=1, rowspan=nrows - 1)

    # Plot heatmap and colorbar
    df_ = df_masked if mask_below_max else df
    hm = sns.heatmap(
        df_,
        square=False,
        cbar_ax=ax2,
        ax=ax1,
        cmap="Reds",
        xticklabels=df.columns,
        yticklabels=set_id_names,
        vmin=0,
        vmax=1,
        annot=False,
    )
    hm.set_facecolor("lightgrey")

    # Add group brackets
    for g_idx, group in enumerate(group_labels):
        n_cols_group = group_positions[g_idx][1] - group_positions[g_idx][0] + 1
        ax = plt.subplot2grid((nrows, ncols), (0, group_positions[g_idx][0]), colspan=n_cols_group, rowspan=1)

        ax.axis("off")
        ax.set_xlim([0, n_cols_group])
        ax.set_ylim([0, 1])
        x1, x2 = [0.2, n_cols_group - 0.2]
        y, h = [0, 0.6]  # [0,0.12]#[0,0.3]
        plt.plot([x1, x1, x2, x2], [y, y + h, y + h, y], lw=2.0 * size_factor, c="black")
        plt.text(
            (x1 + x2) * 0.5, (y + h) * 1.5, group, ha="center", va="bottom", color="black", rotation=90
        )  # , fontsize=fsize)

    # Set axes labels
    # ax1.set_xlabel("marker gene")
    ax2.set_ylabel("max. correlation\nwith marker gene")

    # Show and save
    if save:
        fig.savefig(save, bbox_inches="tight", transparent=True)
    plt.show()


def _lineplot(
    selections_info: pd.DataFrame,
    data: Optional[Dict[str, pd.DataFrame]] = None,
    groupby: Optional[str] = None,
    interpolate: bool = True,
    title: Optional[str] = None,
    xlabel: str = None,
    ylabel: str = None,
    figsize: Tuple[float, float] = (8, 5),
    fontsize: int = 18,
    show: bool = True,
    save: Optional[str] = None,
) -> None:
    """Plot interpolated lineplot

    Args:
        selections_info:
            Information on selections for plotting. The dataframe includes:

                - selection ids or alternative names as index
                - column: `path` (mandatory if ``data=None``): path to results csv of each selection (which contains
                  number of clusters (as index) and one column containing the data to plot)
                - optional columns:

                    - `color`: matplotlib color
                    - `linewidth`: matplotlib linewidth
                    - `linestyle`: matplotlib linestyle
                    - `<groupby>`: some annotation that can be used to group the legend

            Note that the legend order will follow the row order in :attr:`selections_info`.
        data:
            Dictionary with dataframes containing the data to plot for each selection. The keys need to be the same as
            the index of ``selections_info``.
        groupby:
            Column in ``selections_info`` to group the legend.
        interpolate
            Whether to interpolate the values.
        title:
            Plot title.
        xlabel
            X-axis label.
        ylabel:
            Y-axis label.
        figsize:
            Matplotlib figsize.
        fontsize:
            Matplotlib fontsize.
        show:
            Whether to display the plot.
        save:
            Save the plot to path.

    """

    df = selections_info.copy()

    # check if data is there
    if data is None and "path" not in df:
        raise ValueError("No data provided.")

    if data is None:
        data = {}
    else:
        data = {selection_id: data[selection_id] for selection_id in df.index}

    # load data from files if necessary
    if "path" in df:
        for selection_id, path in zip(df.index, df["path"]):
            data[selection_id] = pd.read_csv(path, index_col=0)

    if groupby and groupby not in df:
        raise ValueError(f"To group by {groupby}, a correspondant column is necessary in selections_info.")

    # check style options
    for col in ["color", "linewidth", "linestyle"]:
        # if not given, use default matplotlib values except color if grouping
        if col not in df.columns:
            df[col] = None
        if col == "color" and groupby:
            # if necessary, set grouped line colors (note that just like matplotlib, we cycle over a limited number of
            # (default 10) colors)
            cycler = matplotlib.rcParams["axes.prop_cycle"]
            prop_cycler = itertools.cycle(cycler)
            for group in df[groupby].drop_duplicates():
                df["color"][df[groupby] == group] = next(prop_cycler)["color"]

    # assert that styles are equal for selections that belong to one group
    if groupby:
        group_check = df[["color", "linewidth", "linestyle", groupby]].groupby(by=groupby).nunique() > 1
        if group_check.any().any():
            bad_group = group_check[group_check].dropna(how="all").dropna(how="all", axis=1)
            raise ValueError(
                "Grouping by "
                + groupby
                + " failed because the values of "
                + str(bad_group.columns.values)
                + " are not unique for group(s) "
                + str(bad_group.index.values)
                + "."
            )

    fig = plt.figure(figsize=figsize)

    for selection_id in df.index:

        if selection_id not in data:
            raise ValueError(f"Can't find {selection_id} in selections_info.")

        xlabel = data[selection_id].index.name if not xlabel else xlabel
        yname = data[selection_id].columns[0]
        ylabel = yname if not ylabel else ylabel

        plot_df = data[selection_id]
        label = selection_id if not groupby else df.loc[selection_id][groupby]

        plt.plot(
            plot_df[yname].interpolate() if interpolate else plot_df[yname],
            c=df.loc[selection_id]["color"],
            lw=df.loc[selection_id]["linewidth"],
            linestyle=df.loc[selection_id]["linestyle"],
            label=label,
        )

    if title:
        plt.title()

    plt.xlabel(xlabel, fontsize=fontsize)
    plt.ylabel(ylabel, fontsize=fontsize)
    plt.legend(loc="center left", bbox_to_anchor=(1, 0.5), frameon=False, fontsize=fontsize)
    plt.tick_params(axis="both", labelsize=fontsize)

    # plt.tight_layout()
    if show:
        plt.show()
    if save:
        fig.savefig(save, bbox_inches="tight", transparent=True)
    plt.close()


def knn_overlap(
    selections_info: pd.DataFrame,
    data: Optional[Dict[str, pd.DataFrame]] = None,
    groupby: Optional[str] = None,
    interpolate: bool = True,
    figsize: Tuple[float, float] = (8, 5),
    fontsize: int = 18,
    show: bool = True,
    save: Optional[str] = None,
):
    """Plot mean knn overlap over k

    Args:
        selections_info:
            Information on selections for plotting. The dataframe includes:

                - selection ids or alternative names as index
                - column: `path` (mandatory if ``data=None``): path to results csv of each selection (which contains
                  number of clusters (as index) and one column containing the data to plot)
                - optional columns:

                    - `color`: matplotlib color
                    - `linewidth`: matplotlib linewidth
                    - `linestyle`: matplotlib linestyle
                    - `<groupby>`: some annotation that can be used to group the legend

            Note that the legend order will follow the row order in :attr:`selections_info`.
        data:
            Dictionary with dataframes containing the data to plot for each selection. The keys need to be the same as
            the index of ``selections_info``.
        groupby:
            Column in ``selections_info`` to group the legend.
        interpolate
            Whether to interpolate the values.
        figsize:
            Matplotlib figsize.
        fontsize:
            Matplotlib fontsize.
        show:
            Whether to display the plot.
        save:
            Save the plot to path.

    """

    _lineplot(
        selections_info=selections_info,
        data=data,
        groupby=groupby,
        interpolate=interpolate,
        title=None,
        xlabel="number of neighbors",
        ylabel="mean knn overlap",
        figsize=figsize,
        fontsize=fontsize,
        show=show,
        save=save,
    )


def cluster_similarity(
    selections_info: pd.DataFrame,
    data: Optional[Dict[str, pd.DataFrame]] = None,
    groupby: Optional[str] = None,
    interpolate: bool = True,
    figsize: Tuple[float, float] = (8, 5),
    fontsize: int = 18,
    show: bool = True,
    save: Optional[str] = None,
):
    """Plot cluster similarity as NMI over number of clusters

    Args:
        selections_info:
            Information on selections for plotting. The dataframe includes:

                - selection ids or alternative names as index
                - column: `path` (mandatory if ``data=None``): path to results csv of each selection (which contains
                  number of clusters (as index) and one column containing the data to plot)
                - optional columns:

                    - `color`: matplotlib color
                    - `linewidth`: matplotlib linewidth
                    - `linestyle`: matplotlib linestyle
                    - `<groupby>`: some annotation that can be used to group the legend

            Note that the legend order will follow the row order in :attr:`selections_info`.
        data:
            Dictionary with dataframes containing the data to plot for each selection. The keys need to be the same as
            the index of ``selections_info``.
        groupby:
            Column in ``selections_info`` to group the legend.
        interpolate
            Whether to interpolate the values.
        figsize:
            Matplotlib figsize.
        fontsize:
            Matplotlib fontsize.
        show:
            Whether to display the plot.
        save:
            Save the plot to path.

    """

    _lineplot(
        selections_info=selections_info,
        data=data,
        groupby=groupby,
        interpolate=interpolate,
        title=None,
        xlabel="number of clusters",
        ylabel="NMI",
        figsize=figsize,
        fontsize=fontsize,
        show=show,
        save=save,
    )


def summary_table(
    table: pd.DataFrame,
    summaries: Union[Literal["all"], List[str]] = "all",
    color_maps: Dict[str, matplotlib.colors.Colormap] = {},
    rename_cols: Dict[str, str] = {},
    rename_rows: Dict[str, str] = {},
    time_format: List[str] = [],
    log_scale: List[str] = [],
    color_limits: Dict[str, List[float]] = {},
    nan_color: str = "lightgrey",
    threshold_ann: Dict = {},
    show: bool = True,
    save: Union[bool, str] = False,
) -> None:
    """Plot table of summary statistics

    Args:
        table:
            Dataframe with set ids in the index and a metric in each column.
        summaries:
            List of summary metrics that are plotted.
        color_maps:
            Color maps assigned to summary metrics. Use the initial name and not the potential new name
            given via `rename_cols`.
        rename_cols:
            Rename summary metrics for plot.
        rename_rows:
            Rename set ids.
        time_format:
            Summary names that are formatted to days, hours, mins and secs (seconds are expected as input).
        log_scale:
            Summary names for which a log scaled colormap is applied.
        color_limits:
            For each summary metric optionally provide vmin and vmax for the colormap.
        nan_color:
            Color for nan values.
        threshold_ann:
            Special annotations for values above/below defined thresholds. E.g.
            ``{"time":{"th":1000,"above":True,"ann":"> 1k"}}``
        show:
            Show the figure.
        save:
            Save the plot to path.
    """

    fsize = 15

    # Default order and colors
    default_order = [
        "cluster_similarity",
        "knn_overlap",
        "Greens",
        "forest_clfs",
        "marker_corr",
        "gene_corr",
        "penalty",
    ]
    default_cmaps = {
        "cluster_similarity": "Greens",
        "knn_overlap": "Greens",
        "forest_clfs": "Purples",  # "Reds",
        "marker_corr": "Purples",  # "Reds",
        "gene_corr": "Blues",
        "penalty": truncate_colormap(plt.get_cmap("Greys"), minval=0.05, maxval=0.7, n=100),  # "Greys",
        "other": "Greys",
    }
    if summaries == "all":
        summaries = table.columns.tolist()
        assert isinstance(summaries, list)
        for s in summaries:
            if s not in default_order:
                default_order.append(s.split()[0])
        # Order by default order of metrics and length of summary
        summaries.sort(key=lambda s: default_order.index(s.split()[0]) * 100 + len(s))

    cmaps = {}
    for summary in summaries:
        if summary in color_maps:
            cmaps[summary] = color_maps[summary]
        elif summary.split()[0] in default_cmaps:
            cmaps[summary] = default_cmaps[summary.split()[0]]
        else:
            cmaps[summary] = default_cmaps["other"]

    # Init final table for plotting
    df = table[summaries].copy()

    # Register potential new names of columns that are time formatted or log transformed
    for col in df.columns:
        if (col in time_format) and (col in rename_cols):
            time_format.append(rename_cols[col])
        if (col in log_scale) and (col in rename_cols):
            log_scale.append(rename_cols[col])
        if (col in color_limits) and (col in rename_cols):
            color_limits[rename_cols[col]] = color_limits[col]
        if (col in threshold_ann) and (col in rename_cols):
            threshold_ann[rename_cols[col]] = threshold_ann[col]

    # Rename columns
    df = df.rename(columns=rename_cols, index=rename_rows)

    # Replace old column names with new names in colormaps
    for summary, new_key in rename_cols.items():
        cmaps[new_key] = cmaps.pop(summary)

    n_cols = len(df.columns)
    n_sets = len(df.index)

    fig = plt.figure(figsize=(n_cols * 1.1, n_sets))
    gs1 = GridSpec(1, n_cols)
    gs1.update(wspace=0.0, hspace=0.0)

    multi_col = {}
    cols = df.columns.tolist()
    for col in df.columns.unique():
        count = cols.count(col)
        if count > 1:
            multi_col[col] = 0

    for i, col in enumerate(df.columns):

        ax = plt.subplot(gs1[i])

        yticklabels = bool(i == 0)

        if col in multi_col:  # TODO: time formating multi col support? Better get col pos at the beginning + iloc
            col_pos = [i for i, c in enumerate(df.columns) if c == col][multi_col[col]]
            color_vals = np.log(df.iloc[:, [col_pos]]) if (col in log_scale) else df.iloc[:, [col_pos]]
            multi_col[col] += 1
        else:
            color_vals = np.log(df[[col]]) if (col in log_scale) else df[[col]]
            col_pos = [i for i, c in enumerate(df.columns) if c == col][0]

        if col in time_format:
            # annot = df[col].apply(format_time).values[:, np.newaxis]
            annot = df.iloc[:, col_pos].apply(format_time).values[:, np.newaxis]
            fmt = ""
        else:
            annot = True
            fmt = ".2f"
        if col in threshold_ann:
            formatter = lambda s: f"{s:.2f}"
            annot = df.iloc[:, col_pos].apply(formatter).values[:, np.newaxis] if isinstance(annot, bool) else annot
            tmp = threshold_ann[col]
            th_mask = (df.iloc[:, col_pos] > tmp["th"]) if tmp["above"] else (df.iloc[:, col_pos] < tmp["th"])
            annot[th_mask, :] = tmp["ann"]
            fmt = ""

        g = sns.heatmap(
            color_vals,
            cmap=cmaps[col],
            annot=annot,
            mask=color_vals.isnull(),
            cbar=False,
            square=True,
            yticklabels=yticklabels,
            fmt=fmt,
            annot_kws={"fontsize": fsize - 2},
            vmin=color_limits[col][0] if (col in color_limits) else None,
            vmax=color_limits[col][1] if (col in color_limits) else None,
        )
        g.set_facecolor(nan_color)
        plt.tick_params(
            axis="x", which="major", labelsize=fsize, labelbottom=False, bottom=False, top=True, labeltop=True
        )
        plt.tick_params(axis="y", which="major", labelsize=fsize)
        ax.set_xticklabels(ax.get_xticklabels(), horizontalalignment="left", rotation=45)
        ax.set_yticklabels(ax.get_yticklabels(), rotation=0)
    if show:
        plt.show()
    if save:
        fig.savefig(save, bbox_inches="tight", transparent=True)
    plt.close()


#############################
## selection related plots ##
#############################


def explore_constraint(
    a: List[sc.AnnData],
    selections_tmp: List[pd.DataFrame],
    penalty_kernels: List[Callable],
    x_axis_key: str = "quantile_99",
    factors: List[float] = None,
    upper: float = 6,
    lower: float = 2,
    size_factor: int = 6,
    n_rows: int = 1,
    legend_size: int = 9,
    show: bool = True,
    save: Optional[str] = None,
) -> None:
    """Plot histogram of quantiles for selected genes for different penalty kernels.

    Args:
        a:
            List of ``sc.AnnData`` objects containing the data for plotting.
        selections_tmp:
            Dataframe containing the selection.
        factors:
            List of titles for the subplots, i.e. the factors of each penalty kernel.
        upper:
            Lower border above which the kernel is 1.
        lower:
            Upper boder below which the kernel is 1.
        penalty_kernels:
            List of penalty kernels, which were used for selection.
        x_axis_key:
            Key of column in ``adata.var`` that is used for the x axis of the plotted histograms.
        size_factor:
             Factor for scaling the figure size.
        n_rows:
            Number of subplot rows.
        legend_size:
            Matplotlib legend size.
        show:
            Whether to display the plot.
        save:
            Save the plot to path.

    Returns:

    """

    # TODO:
    #  remove --> now selection_histogram
    #  1) Fix explore_constraint plot. The following circular import is causing problems atm:
    #     DONE: moved parts of this method to selector.plot_expore_constraint --> this solves the problem
    #  2) How to generalize the plotting function, support:
    #     - any selection method with defined hyperparameters
    #     - any penalty kernel
    #     - any key to be plotted (not only quantiles)

    if factors is None:
        factors = [1]
    assert isinstance(factors, list)

    cols = len(factors)

    fig = plt.figure(figsize=(size_factor * cols, 0.7 * size_factor * n_rows))
    for i, factor in enumerate(factors):
        ax1 = plt.subplot(n_rows, cols, i + 1)
        hist_kws = {"range": (0, np.max(a[i].var[x_axis_key]))}
        bins = 100
        sns.distplot(
            a[i].var[x_axis_key],
            kde=False,
            label="highly_var",
            bins=bins,
            hist_kws=hist_kws,
        )
        idx = selections_tmp[i]["selection"].index[selections_tmp[i]["selection"]]
        sns.distplot(
            a[i][:, idx].var[x_axis_key],
            kde=False,
            label="selection",
            bins=bins,
            hist_kws=hist_kws,
        )
        plt.axvline(x=lower, lw=0.5, ls="--", color="black")
        plt.axvline(x=upper, lw=0.5, ls="--", color="black")
        ax1.set_yscale("log")
        plt.legend(prop={"size": legend_size}, loc=[0.73, 0.74], frameon=False)
        plt.title(f"factor = {factor}")

        ax2 = ax1.twinx()
        x_values = np.linspace(0, np.max(a[i].var[x_axis_key]), 240)
        plt.plot(x_values, 1 * penalty_kernels[i](x_values), label="penal.", color="green")
        plt.legend(prop={"size": legend_size}, loc=[0.73, 0.86], frameon=False)
        plt.ylim([0, 2])
        for label in ax2.get_yticklabels():
            label.set_color("green")

    plt.tight_layout()
    if show:
        plt.show()
    if save:
        fig.savefig(save, bbox_inches="tight", transparent=True)
    plt.close()


def selection_histogram(
    adata: sc.AnnData,
    selections_dict: Dict[str, pd.DataFrame],
    x_axis_keys: Dict[str, str],
    background_key: str = "highly_variable",
    penalty_kernels: Dict[str, Dict[str, Callable]] = None,
    penalty_keys: Dict[str, List[str]] = None,
    penalty_labels: Union[str, Dict[str, Dict[str, str]]] = "penalty",
    upper_borders: Union[bool, Dict[str, Union[float, bool]]] = None,
    lower_borders: Union[bool, Dict[str, Union[float, bool]]] = None,
    size_factor: float = 4.5,
    legend_size: float = 11,
    show: bool = True,
    save: Optional[str] = None,
) -> None:
    """Plot histogram of quantiles for selected genes for different penalty kernels.

    Args:
        adata:
            ``sc.AnnData`` object containing the data for plotting the histogram in ``adata.var[x_axis_key]``.
        selections_dict:
            Dictionary of dataframes with gene identifiers as index and a boolean column indicating the selection. The
            dictionary keys are used as label in the plot and need to be equivalent to the keys in ``penalty_kernels``
            and ``penalty_keys``.
        x_axis_keys:
            Dictionary with the column in ``adata.var`` that is used for the x axis of the plotted histograms. There are
            two reasonable options:

                - If a penalty is plottet: the column containing the values of which the penalty scores were derived by
                  applying the penalty kernel. Use the ``penalty_key`` as dictionary key in this case.
                - If no penalty is plottet: a column containing some statistic, eg. 99% quantile. Use the same keys as
                  ``selection_dict`` in this case.

        background_key:
            Key of column in ``adata.var`` which is plottet as background histogram. If `None`, no background histogram
            is plotted. If `'all'`, all genes are used as background.
        penalty_kernels:
            Dictionary of penalty kernels, which were used for each selection. The outer key is the selection name. The
            inner keys are used as subplot titles.
            Additionally, ``penalty_keys`` can be provided. If both are None, only the histograms are plottet.
        penalty_keys:
            Dictionary of a list of column keys of ``adata.var`` containing penalty scores for each selection.
            Additionally, ``penalty_kernels`` can be provided. If both are None, only the histograms are plottet.
        penalty_labels:
            A legeng label for each selection and each penalty. The keys of the outer dictionary need to be the
            selection names. As keys for the inner dictionary, use the penalty keys.
        upper_borders:
            Dictionary with the lower borders above which the kernels are 1, which is indicated as a vertical line in
            the plot. Use the same dictionay keys as ``penalty_keys`` or ``penalty_kernels``. If None, it is
            automatically infered. If False or not in the dictionary, no vertical line at the upper border is drawn.
        lower_borders:
            Dictionary with the upper borders below which the kernels are 1, which is indicated as a vertical line in
            the plot. Use the same dictionay keys as ``penalty_keys`` or ``penalty_kernels``. If None, it is
            automatically infered. If False or not in the dictionary, no vertical line at the lower border is drawn.
        size_factor:
             Factor for scaling the figure size.
        legend_size:
            Matplotlib legend size.
        show:
            Whether to display the plot.
        save:
            Save the plot to path.

    Returns:

    """

    x_values_dict = {}
    y_values_dict = {}

    if penalty_labels is None:
        penalty_labels = {}

    # apply penalty kernel if given
    if penalty_kernels is not None:
        for selection_label in selections_dict:

            x_values_dict[selection_label] = {}
            y_values_dict[selection_label] = {}

            if selection_label in penalty_kernels:
                for i, penalty_key in enumerate(penalty_kernels[selection_label]):

                    penalty_kernel = penalty_kernels[selection_label][penalty_key]

                    if penalty_key not in x_axis_keys:
                        raise ValueError(f"No adata.var key given for {penalty_key}.")
                    x_values = np.linspace(0, np.max(adata.var[x_axis_keys[penalty_key]]), 240)

                    x_values_dict[selection_label][penalty_key] = x_values
                    y_values_dict[selection_label][penalty_key] = penalty_kernel(x_values)

    # interpolate penalty scores if values in adata.var[[penalty_keys]] are given
    if penalty_keys is not None:
        for selection_label in selections_dict:

            x_values_dict[selection_label] = {}
            y_values_dict[selection_label] = {}

            if selection_label in penalty_keys:
                for i, penalty_key in enumerate(penalty_keys[selection_label]):

                    if penalty_key not in adata.var:
                        raise ValueError(f"Can't plot {penalty_key} because it was not found in adata.var. ")

                    if penalty_key not in x_axis_keys:
                        raise ValueError(f"No adata.var key given for {penalty_key}.")

                    penalty_interp = interp1d(
                        adata.var[x_axis_keys[penalty_key]], adata.var[penalty_key], kind="linear"
                    )
                    x_values = np.linspace(penalty_interp.x[0], penalty_interp.x[-1], 240)

                    x_values_dict[selection_label][penalty_key] = x_values
                    y_values_dict[selection_label][penalty_key] = penalty_interp(x_values)

    if (penalty_kernels is None) and (penalty_keys is None):
        x_values_dict = {set_id: [] for set_id in selections_dict}

    n_rows = len(selections_dict)
    cols = [len(x_values_dict[x]) for x in x_values_dict]
    n_cols = max(max(cols), 1)  # if cols else 1

    fig = plt.figure(figsize=(size_factor * n_cols, 0.7 * size_factor * n_rows))
    gs = GridSpec(n_rows, n_cols, figure=fig)
    i = -1
    for selection_label, selection_df in selections_dict.items():

        for j in range(n_cols):
            i = i + 1
            # if no penalty is given, plot only one histogram without penalty
            if len(x_values_dict[selection_label]) == 0 and j > 0:
                continue
            # if less penalties given than n_cols, let space empty
            elif 0 < len(x_values_dict[selection_label]) <= j:
                continue
            ax1 = fig.add_subplot(gs[i])

            # get adata.var key for histogram
            if len(x_values_dict[selection_label]) == 0:
                # no penalty to plot
                var_key = x_axis_keys[selection_label]
            else:
                # penalty kernel should be plottet
                penalty_key = list(x_values_dict[selection_label].keys())[j]
                var_key = x_axis_keys[penalty_key]
            assert isinstance(var_key, str)

            # get histogram data
            selected_genes = selection_df.loc[selection_df["selection"]].index  # selection_df.index[selection_df]
            mask = adata.var.index.isin(selected_genes)
            hist_data = adata[:, mask].var[var_key]
            hist_kws = {"range": (0, np.max(hist_data))}

            # get background data
            bg_data = None
            if background_key == "all":
                bg_data = adata.var[var_key]
            elif background_key:
                bg_data = adata[:, adata.var[background_key]].var[var_key]

            # plot background histogram
            if bg_data is not None:
                hist_kws = {"range": (0, np.max(bg_data))}
                bins = 100
                sns.distplot(
                    bg_data,
                    kde=False,
                    label=background_key,
                    bins=bins,
                    hist_kws=hist_kws,
                )

            # plot selection histogram
            bins = 100
            sns.distplot(
                hist_data,
                kde=False,
                label=selection_label,
                bins=bins,
                hist_kws=hist_kws,
            )
            ax1.set_yscale("log")
            if j == 0:
                ax1.set_ylabel("number of genes")

            # check if penalty is there to plot
            if len(x_values_dict[selection_label]) <= j:
                plt.legend(prop={"size": legend_size}, bbox_to_anchor=[0.99, 0.99], loc="upper right", frameon=False)
                continue
            assert isinstance(penalty_key, str)

            # draw vertical lines at lower and upper border
            ax2 = ax1.twinx()
            if lower_borders is None:
                lower = None
            else:
                if lower_borders is False:
                    lower = False
                else:
                    assert isinstance(lower_borders, dict)
                    if penalty_key not in lower_borders:
                        lower = False

            if lower is None:
                # infer lower border from penanalty values
                idx_of_first_one = np.where(y_values_dict[selection_label][penalty_key] >= 0.999)[0][0]
                lower = x_values_dict[selection_label][penalty_key][idx_of_first_one] if idx_of_first_one > 0 else False
            if lower is not False:
                assert isinstance(lower, float)
                plt.axvline(x=lower, lw=0.5, ls="--", color="black")

            if upper_borders is None:
                upper = None
            else:
                if upper_borders is False:
                    upper = False
                else:
                    assert isinstance(upper_borders, dict)
                    if penalty_key not in upper_borders:
                        upper = False

            if upper is None:
                # infer lower border from penanalty values
                idx_of_last_one = np.where(y_values_dict[selection_label][penalty_key] > 0.999)[0][-1]
                upper = (
                    x_values_dict[selection_label][penalty_key][idx_of_last_one]
                    if idx_of_last_one < len(x_values_dict[selection_label][penalty_key]) - 1
                    else False
                )
            if upper is not False:
                assert isinstance(upper, float)
                plt.axvline(x=upper, lw=0.5, ls="--", color="black")

            # plot penalty kernel (interpolated if from penalty_key)
            plt.title(penalty_key)
            penalty_label = "penalty"
            if selection_label in penalty_labels:
                if penalty_key in penalty_labels[selection_label]:
                    penalty_label = penalty_labels[selection_label][penalty_key]
            plt.plot(
                x_values_dict[selection_label][penalty_key],
                y_values_dict[selection_label][penalty_key],
                label=penalty_label,
                color="green",
            )

            h1, l1 = ax1.get_legend_handles_labels()
            h2, l2 = ax2.get_legend_handles_labels()
            plt.legend(
                h1 + h2,
                l1 + l2,
                prop={"size": legend_size},
                bbox_to_anchor=[0.99, 0.99],
                loc="upper right",
                frameon=False,
            )

            plt.ylim([0, 2])
            for label in ax2.get_yticklabels():
                label.set_color("green")

    plt.tight_layout()
    if show:
        plt.show()
    if save:
        fig.savefig(save, bbox_inches="tight", transparent=True)
    plt.close()


def gene_overlap(
    selection_df: pd.DataFrame,
    style: Literal["upset", "venn"] = "upset",
    min_degree: int = 0,
    fontsize: int = 18,
    show: bool = True,
    save: Optional[str] = None,
) -> None:
    """Plot overlap of gene sets

    Args:
        selection_df:
            Table with gene sets. Gene names are given in the index, gene sets are given as boolean columns.
        style:
            Plot type. Options are

                - "upset": upset plot
                - "venn": venn diagram

        min_degree:
            Only for `style="upset"`: minimum degree (number of categories intersected) of a subset to be shown in the
            plot.
        fontsize:
            Matplotlib fontsize.
        show:
            Whether to display the plot.
        save:
            Save the plot to path.

    """

    if style == "venn":
        # remove emtpy sets:
        selection_df = selection_df.loc[:, selection_df.sum() > 0]

        # calculate plotting parameter
        labels, radii, actualOverlaps, disjointOverlaps = venn.df2areas(selection_df, fineTune=False)

        # draw venn diagram
        fig, ax = venn.venn(
            radii, actualOverlaps, disjointOverlaps, labels=labels, labelsize=fontsize, cmap="Blues", fineTune=False
        )

    elif style == "upset":
        # transform to compatible format
        upset_data = from_indicators(selection_df)

        # set up figure
        upset_plot = UpSet(
            upset_data, subset_size="count", min_degree=min_degree, show_counts=True, sort_by="cardinality"
        )

        # draw figure
        fig = plt.figure()
        upset_plot.plot(fig=fig)

    plt.tight_layout()
    if show:
        plt.show()
    if save:
        fig.savefig(save, bbox_inches="tight", transparent=True)
    plt.close()


# def gene_overlap_grouped(
#     selection_df: Dict[str, pd.DataFrame],
#     groupby: str = "method",
#     show: bool = True,
#     save: Optional[str] = None,
# ):
#     """Plot the intersection of different selected gene sets grouped by the selection method.
#
#     Args:
#         selection_df:
#             Boolean dataframe with gene identifiers as index and one column for each gene set.
#         groupby:
#             Name of a column that categorizes the gene sets, eg. the method they were selected with.
#         show:
#             Whether to display the plot.
#         save:
#             Save the plot to path.
#
#     Returns:
#         Figure can be shown (default `True`) and stored to path (default `None`).
#         Change this with ``show`` and ``save``.
#
#     """
#
#     pass
#     # TODO


def clf_genes_umaps(
    adata: sc.AnnData,  # TODO: adjust type hint
    df: pd.DataFrame,
    basis: str = "X_umap",
    ct_key: str = "celltype",
    n_cols: int = 4,
    size_factor: float = 1,
    fontsize: int = 18,
    show: bool = True,
    save: Optional[str] = None,
) -> None:
    """Plot umaps of genes needed for cell type classification of each cell type.

    Args:
        adata:
            ``AnnData`` with ``ct_key`` in ``adata.obs`` and ``basis`` in ``adata.obsm``.
        df:
            Dataframe with genes as index and the following columns:

                - `'decision_celltypes'`: List of celltypes for which the gene is needed for classification.
                - `'rank'`: Forest classification rank of the gene.
                - `'importance_score'`: Importance score of the gene.

            Optional columns: TODO: do we need to list them here?
                - `'marker_celltypes'`: List of celltypes for which the gene is a marker according to the known marker
                    list. TODO: check if also DE "markers" alowed?
                - `'decision_title'`: Subplot title.
                - `'marker_title'`: Subplot title used if gene is marker. TODO: why decision_title AND marker_title?
                - `'decision_cmap'`: Matplotlib colormap.
                - `'marker_cmap'`: Matplotlib colormap used if gene is marker. TODO: why decision_cmap and marker_cmap (and why cmap at all...)

        basis:
            Name of the ``obsm`` embedding to use.
        ct_key:
            Column name in ``adata.obs`` where celltypes are stored.
        n_cols:
            Number of subplot columns.
        fontsize:
            Matplotlib fontsize.
        size_factor:
            Scale factor for figure width and height.
        show:
            Whether to display the plot.
        save:
            Save the plot to path.

    Returns:
        Figure can be shown (default `True`) and stored to path (default `None`).
        Change this with ``show`` and ``save``.
    """

    # prepare data
    a = adata.copy()
    celltypes = list(set([y for x in df["decision_celltypes"] for y in x]))
    subplots_decision = {ct: list(df.index[df["decision_celltypes"].apply(lambda x: ct in x)]) for ct in celltypes}
    subplots_marker = {ct: [] for ct in celltypes}
    if "marker_celltypes" in df:
        subplots_marker = {ct: list(df.index[df["marker_celltypes"] == ct]) for ct in celltypes}

    # set titles and palettes:
    if "decision_title" not in df:
        df["decision_title"] = [
            f"{gene}: \n" f"rank={int(df.loc[gene]['rank'])}, " f"imp.={round(df.loc[gene]['importance_score'], 2)}"
            for gene in df.index
        ]
    if "marker_title" not in df:
        df["marker_title"] = [
            f"marker: {gene}: \n"
            f"rank={int(df.loc[gene]['rank'])}, "
            f"imp.={round(df.loc[gene]['importance_score'], 2)}"
            for gene in df.index
        ]
    if "decision_palette" not in df:
        # cmap = colors.LinearSegmentedColormap.from_list("mycmap", ["grey", "lime"])
        df["decision_cmap"] = "viridis"  # cmap
    if "marker_palette" not in df:
        # cmap = colors.LinearSegmentedColormap.from_list("mycmap", ["grey", "orangered"])
        df["marker_cmap"] = "viridis"  # cmap

    # numbers of subplots in rows and columns
    n_decision_genes = [len(subplots_decision[ct] + subplots_marker[ct]) for ct in subplots_decision]
    n_subplots = [n + 1 for n in n_decision_genes]
    if n_cols is None:
        n_cols = max(n_subplots)
    rows_per_ct = [np.ceil(s / n_cols) for s in n_subplots]
    n_rows = int(sum(rows_per_ct))
    row_ceils = [int(np.ceil(s / r)) for s, r in zip(n_subplots, rows_per_ct)]
    n_cols = max(row_ceils)

    CT_FONTSIZE = fontsize + 4
    PPI = 72
    CT_PADDING = CT_FONTSIZE / PPI  # space above celltype (additional to HSPACE)

    HSPACE_INCHES = fontsize / PPI * n_rows * 3.5
    WSPACE_INCHES = fontsize / PPI * n_cols * 4
    TOP_INCHES = -CT_PADDING
    BOTTOM_INCHES = 3
    LEFT_INCHES = 3
    RIGHT_INCHES = 3
    CT_HEIGHT_INCHES = CT_PADDING + (CT_FONTSIZE / PPI)
    SUBPLOT_HEIGHT_INCHES = 3
    SUBPLOT_WIDTH_INCHES = 3

    FIGURE_WIDTH = (
        (SUBPLOT_WIDTH_INCHES * n_cols) + (((n_cols - 1) / n_cols) * WSPACE_INCHES) + RIGHT_INCHES + LEFT_INCHES
    ) * size_factor
    FIGURE_HEIGHT = (
        ((SUBPLOT_HEIGHT_INCHES + CT_HEIGHT_INCHES) * n_rows)
        + (((n_rows - 1) / n_rows) * HSPACE_INCHES)
        + TOP_INCHES
        + BOTTOM_INCHES
    ) * size_factor

    HSPACE = HSPACE_INCHES / FIGURE_HEIGHT
    WSPACE = WSPACE_INCHES / FIGURE_WIDTH
    TOP = 1 - (TOP_INCHES / FIGURE_HEIGHT)
    BOTTOM = BOTTOM_INCHES / FIGURE_HEIGHT
    RIGHT = 1 - (RIGHT_INCHES / FIGURE_WIDTH)
    LEFT = LEFT_INCHES / FIGURE_WIDTH
    SUBPLOT_HEIGHT = SUBPLOT_HEIGHT_INCHES / FIGURE_HEIGHT
    CT_HEIGHT = CT_HEIGHT_INCHES / FIGURE_HEIGHT

    fig = plt.figure(figsize=(FIGURE_WIDTH, FIGURE_HEIGHT))

    # note that every second row is just for legends
    gs = GridSpec(n_rows * 2, n_cols, figure=fig, height_ratios=[CT_HEIGHT, SUBPLOT_HEIGHT] * n_rows)
    i = -1
    for ct in celltypes:
        i += 1
        j = 0

        # prepare data
        a.obs[ct] = adata.obs[ct_key].astype(str)
        a.obs.loc[a.obs[ct] != ct, ct] = "other"
        a.obs[ct] = a.obs[ct].astype("category")

        # first subplot is the umap colored by celltype
        ax = fig.add_subplot(gs[2 * i + 1, j])
        ax = sc.pl.embedding(
            adata=a,
            basis=basis,
            color=ct,
            show=False,
            ax=ax,
            title="",
            palette=["blue", "grey"],
            legend_fontweight="heavy",
        )
        ax.xaxis.label.set_fontsize(fontsize)
        ax.yaxis.label.set_fontsize(fontsize)
        ax.title.set_fontsize(fontsize)
        ax.get_legend().remove()

        # add celltype header
        celltype_ax = fig.add_subplot(gs[2 * i, j])
        celltype_ax.text(
            x=0,
            y=0,
            s=ct,
            weight="bold",
            verticalalignment="baseline",
            horizontalalignment="left",
            fontsize=CT_FONTSIZE,
        )
        celltype_ax.set_axis_off()

        # subplots for decision genes:
        for gene in subplots_decision[ct]:

            j += 1
            if j >= n_cols:
                i += 1
                j = 0

            # set styling
            ax = fig.add_subplot(gs[2 * i + 1, j])
            ax = sc.pl.embedding(
                adata=a,
                basis=basis,
                color=gene,
                show=False,
                ax=ax,
                title=df["decision_title"][gene],
                cmap=df["decision_cmap"][gene],
            )
            ax.xaxis.label.set_fontsize(fontsize)
            ax.yaxis.label.set_fontsize(fontsize)
            ax.title.set_fontsize(fontsize)
            cbar = ax.collections[-1].colorbar
            cbar.ax.tick_params(labelsize=fontsize)

        # subplots for marker genes
        for gene in subplots_marker[ct]:

            j += 1
            if j >= n_cols:
                i += 1
                j = 0

            ax = fig.add_subplot(gs[2 * i + 1, j])
            ax = sc.pl.embedding(
                adata=a,
                basis=basis,
                color=gene,
                show=False,
                ax=ax,
                title=df.loc[gene]["marker_title"],
                cmap=df["marker_cmap"][gene],
            )
            ax.xaxis.label.set_fontsize(fontsize)
            ax.yaxis.label.set_fontsize(fontsize)
            ax.title.set_fontsize(fontsize)
            cbar = ax.collections[-1].colorbar
            cbar.ax.tick_params(labelsize=fontsize)

    plt.subplots_adjust(bottom=BOTTOM, top=TOP, left=LEFT, right=RIGHT, hspace=HSPACE, wspace=WSPACE)
    if show:
        plt.show()
    if save:
        fig.savefig(save, bbox_inches="tight", transparent=True)
    plt.close()


def masked_dotplot(
    adata: sc.AnnData,  # TODO: adjust type hint
    selector,
    ct_key: str = "celltype",
    imp_threshold: float = 0.05,
    celltypes: Optional[List[str]] = None,
    n_genes: Optional[int] = None,
    comb_markers_only: bool = False,
    markers_only: bool = False,
    cmap: str = "Reds",
    comb_marker_color: str = "darkblue",
    marker_color: str = "blue",
    non_adata_celltypes_color: str = "grey",
    save: Union[bool, str] = False,
):
    """Create dotplot with additional annotation masks.

    Args:
        adata:
            AnnData with cell type annotations in `adata.obs[ct_key]`.
        selector:
            `ProbesetSelector` object with selected `selector.probeset`.
        ct_key:
            Column of `adata.obs` with cell type annotation.
        imp_threshold:
            Annotate genes as "Spapros marker" only for those genes with importance > ``imp_threshold``.
        celltypes:
            Optional subset of celltypes (rows of dotplot).
        n_genes:
            Optionally plot top ``n_genes`` genes.
        comb_markers_only:
            Whether to plot only genes that are "Spapros markers" for the plotted cell types. (can be combined with
            markers_only, in that case markers that are not Spapros markers are also shown)
        markers_only:
            Whether to plot only genes that are markers for the plotted cell types. (can be combined with
            ``comb_markers_only``, in that case Spapros markers that are not markers are also shown)
        cmap:
            Colormap of mean expressions.
        comb_marker_color:
            Color for "Spapros markers".
        marker_color:
            Color for marker genes.
        non_adata_celltypes_color:
            Color for celltypes that don't occur in the data set.
        save:
            If `True` or a `str`, save the figure.


    Example:

        (Takes a few minutes to calculate)

        .. code-block:: python

            import spapros as sp
            adata = sp.ut.get_processed_pbmc_data()
            selector = sp.se.ProbesetSelector(adata, "celltype", n=30, verbosity=0)
            selector.select_probeset()

            sp.pl.masked_dotplot(adata,selector)

        .. image:: ../../docs/plot_examples/masked_dotplot.png


    """
    from spapros.selection import ProbesetSelector

    if isinstance(selector, str):
        selector = ProbesetSelector(adata, ct_key, save_dir=selector)
        # TODO: think the last steps of the ProbesetSelector are still not saved..., needs to be fixed.

    # celltypes, possible origins:
    # - adata.obs[ct_key] (could include cts not used for selection)
    # - celltypes for selection (including markers, could include cts which are not in adata.obs[ct_key])
    # --> pool all together... order?

    if celltypes is not None:
        cts = celltypes
        a = adata[adata.obs[ct_key].isin(celltypes)].copy()
        # a.obs[ct_key] = a.obs[ct_key].astype(str).astype("category")
    else:
        # Cell types from adata
        cts = adata.obs[ct_key].unique().tolist()
        # Cell types from marker list only
        if "celltypes_marker" in selector.probeset:
            tmp = []
            for markers_celltypes in selector.probeset["celltypes_marker"].str.split(","):
                tmp += markers_celltypes
            tmp = np.unique(tmp).tolist()
            if "" in tmp:
                tmp.remove("")
            cts += [ct for ct in tmp if ct not in cts]
        a = adata

    # Get selected genes that are also in adata
    selected_genes = [
        g for g in selector.probeset[selector.probeset["selection"]].index.tolist() if g in adata.var_names
    ]

    # Get tree genes
    tree_genes = {}
    assert isinstance(selector.forest_results["forest"], list)
    assert len(selector.forest_results["forest"]) == 3
    assert isinstance(selector.forest_results["forest"][2], dict)
    for ct, importance_tab in selector.forest_results["forest"][2].items():
        if ct in cts:
            tree_genes[ct] = importance_tab["0"].loc[importance_tab["0"] > imp_threshold].index.tolist()
            tree_genes[ct] = [g for g in tree_genes[ct] if g in selected_genes]

    # Get markers
    marker_genes: Dict[str, list] = {ct: [] for ct in (cts)}
    for ct in cts:
        for gene in selector.probeset[selector.probeset["selection"]].index:
            if ct in selector.probeset.loc[gene, "celltypes_marker"].split(",") and (gene in adata.var_names):
                marker_genes[ct].append(gene)
        marker_genes[ct] = [g for g in marker_genes[ct] if g in selected_genes]

    # Optionally subset genes:
    # Subset to combinatorial markers of shown celltypes only
    if comb_markers_only or markers_only:
        allowed_genes = []
        if comb_markers_only:
            allowed_genes += list(itertools.chain(*[tree_genes[ct] for ct in tree_genes.keys()]))
        if markers_only:
            allowed_genes += list(itertools.chain(*[marker_genes[ct] for ct in marker_genes.keys()]))
        selected_genes = [g for g in selected_genes if g in allowed_genes]
    # Subset to show top n_genes only
    if n_genes:
        selected_genes = selected_genes[: min(n_genes, len(selected_genes))]
    # Filter (combinatorial) markers by genes that are not in the selected genes
    for ct in cts:
        marker_genes[ct] = [g for g in marker_genes[ct] if g in selected_genes]
    for ct in tree_genes.keys():
        tree_genes[ct] = [g for g in tree_genes[ct] if g in selected_genes]

    dp = MaskedDotPlot(
        a,
        var_names=selected_genes,
        groupby=ct_key,
        tree_genes=tree_genes,
        marker_genes=marker_genes,
        further_celltypes=[ct for ct in cts if ct not in adata.obs[ct_key].unique()],
        cmap=cmap,
        tree_genes_color=comb_marker_color,
        marker_genes_color=marker_color,
        non_adata_celltypes_color=non_adata_celltypes_color,
    )
    dp.make_figure()
    if save:
        plt.gcf().savefig(save, bbox_inches="tight", transparent=True)


######################
## helper functions ##
######################


def format_time(time: float) -> str:
    """Reformat a time stamp.

    Args:
        time:
            Time in seconds.

    Returns:
        str:
            The formatted time.
    """
    days = int(time // (3600 * 24))
    hours = int(time // 3600)
    mins = int(time // 60)
    secs = int(time // 1)
    unit = ["d", "h", "min", "sec"]
    for t, u in zip([days, hours, mins, secs], unit):
        if t > 0:
            return f"{t} {u}"
    return "0 sec"


def truncate_colormap(
    cmap: matplotlib.colors.Colormap, minval: float = 0.0, maxval: float = 1.0, n: int = 100
) -> matplotlib.colors.Colormap:
    """Truncate a colormap to a given number of colors and an interval.

    Args:
        cmap:
            Colormap name.
        minval:
            Smallest color value.
        maxval:
            Highest color value.
        n:
            Number of colors.
    """
    new_cmap = colors.LinearSegmentedColormap.from_list(
        "trunc({n},{a:.2f},{b:.2f})".format(n=cmap.name, a=minval, b=maxval), cmap(np.linspace(minval, maxval, n))
    )
    return new_cmap



from spapros.plotting._masked_dotplot import MaskedDotPlot
from spapros.plotting.plot import clf_genes_umaps
from spapros.plotting.plot import cluster_similarity
from spapros.plotting.plot import confusion_matrix
from spapros.plotting.plot import correlation_matrix
from spapros.plotting.plot import explore_constraint
from spapros.plotting.plot import gene_overlap
from spapros.plotting.plot import knn_overlap
from spapros.plotting.plot import marker_correlation
from spapros.plotting.plot import masked_dotplot
from spapros.plotting.plot import selection_histogram
from spapros.plotting.plot import summary_table

__all__ = [
    "correlation_matrix",
    "confusion_matrix",
    "summary_table",
    "explore_constraint",
    "masked_dotplot",
    "MaskedDotPlot",
    "gene_overlap",
    "cluster_similarity",
    "knn_overlap",
    "selection_histogram",
    "clf_genes_umaps",
    "marker_correlation",
]



from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
import scanpy as sc
from rich import print
from rich.progress import Progress as Progress
from scipy.sparse import issparse
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import StratifiedKFold
from sklearn.utils.class_weight import compute_sample_weight
from spapros.util.util import clean_adata
from spapros.util.util import cluster_corr
from spapros.util.util import dict_to_table
from spapros.util.util import gene_means
from spapros.util.util import init_progress
from xgboost import XGBClassifier

METRICS_PARAMETERS: Dict[str, Dict] = {
    "cluster_similarity": {
        "ns": [5, 60],
        "AUC_borders": [[5, 20], [21, 60]],
    },
    "knn_overlap": {
        "ks": [5, 10, 15, 20, 25, 30],
    },
    "forest_clfs": {
        "ct_key": "celltype",
        "threshold": 0.8,
    },
    "marker_corr": {
        "marker_list": None,
        "per_celltype": True,
        "per_marker": True,
        "per_celltype_min_mean": None,
        "per_marker_min_mean": 0.025,
    },
    "gene_corr": {"threshold": 0.8},
}


# utility functions to get general infos about metrics #
def get_metric_names() -> List[str]:
    """Get a list of all available metric names."""
    return [metric for metric in METRICS_PARAMETERS]


def get_metric_parameter_names() -> Dict[str, List[str]]:
    """Get all parameter names for each available metric."""
    names = {metric: [param for param in METRICS_PARAMETERS[metric]] for metric in METRICS_PARAMETERS}
    return names


def get_metric_default_parameters() -> Dict[str, Dict]:
    """Get the default metric parameters."""
    return METRICS_PARAMETERS


############################
# General metric functions #
############################
# We define theses to systematically seperate all metrics in three different parts
# 1. shared computations: only need to compute one time for a given dataset
# 2. specific computations: Need to be computed for each probe set
# 3. summary: Simple final computations (per probe set) that aggregate evaluations to summary metrics


def metric_shared_computations(
    adata: sc.AnnData,
    metric: str,
    parameters: Dict = {},
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
) -> pd.DataFrame:
    """Calculate the metric compuations that can be shared between probe sets.

    Args:
        adata:
            An already preprocessed annotated data matrix. Typically we use log normalised data.
        metric:
            The metric to be calculated.
        parameters:
            Parameters for the calculation of the metric.
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
    """
    if metric not in get_metric_names():
        raise ValueError(f"Unsupported metric: {metric}")

    progress, started = init_progress(progress, verbosity, level)

    description = "Computing shared compuations for " + metric + "..."

    if metric == "cluster_similarity":
        return leiden_clusterings(
            adata, parameters["ns"], progress=progress, level=level, verbosity=verbosity, description=description
        )

    elif metric == "knn_overlap":
        return knns(
            adata,
            genes="all",
            ks=parameters["ks"],
            progress=progress,
            level=level,
            verbosity=verbosity,
            description=description,
        )
    elif metric == "forest_clfs":
        pass
    elif metric == "marker_corr":
        return marker_correlation_matrix(
            adata,
            marker_list=parameters["marker_list"],
            progress=progress,
            level=level,
            verbosity=verbosity,
            description=description,
        )
    elif metric == "gene_corr":
        return correlation_matrix(adata, progress=progress, level=level, verbosity=verbosity, description=description)
    else:
        raise ValueError(f"Unsupported metric: {metric}")

    if progress and started:
        progress.stop()


def metric_pre_computations(
    genes: List,
    adata: sc.AnnData,
    metric: str = None,
    parameters: Dict = {},
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
) -> Union[pd.DataFrame, None]:
    """Calculate the metric computations that are independent of the shared results.

    Note:
        If there are no shared results needed at all to calculate a metric, the computations are put in
        `metric_computations`, this is the case for e.g. forest_clfs.

    Args:
        genes:
            The selected genes.
        adata:
            An already preprocessed annotated data matrix. Typically we use log normalised data.
        metric:
            The metric to be calculated.
        parameters:
            Parameters for the calculation of the metric.
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
    """
    if metric not in get_metric_names():
        raise ValueError(f"Unsupported metric: {metric}")

    progress, started = init_progress(progress, verbosity, level)

    description = "Computing pre compuations for " + metric + ".."

    if metric == "cluster_similarity":
        return leiden_clusterings(
            adata[:, genes],
            parameters["ns"],
            progress=progress,
            level=level,
            verbosity=verbosity,
            description=description,
        )
    elif metric == "knn_overlap":
        return knns(
            adata,
            genes=genes,
            ks=parameters["ks"],
            progress=progress,
            level=level,
            verbosity=verbosity,
            description=description,
        )
    elif metric == "forest_clfs":
        return None
    elif metric == "marker_corr":
        return None
    elif metric == "gene_corr":
        return None
    else:
        raise ValueError(f"Unsupported metric: {metric}")

    if progress and started:
        progress.stop()


def metric_computations(
    genes: list,
    adata: sc.AnnData = None,
    metric: str = None,
    shared_results: pd.DataFrame = None,
    pre_results: pd.DataFrame = None,
    parameters: Dict = {},
    n_jobs: int = -1,
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
) -> pd.DataFrame:
    """Compute the probe set specific evaluation metrics.

    Args:
        genes:
            The selected genes.
        adata:
            An already preprocessed annotated data matrix. Typically we use log normalised data.
        metric:
            The metric to be calculated.
        shared_results:
            The results of the metric calculations, that are not probe set specific.
        pre_results:
            The results of the metric calculations, that are independent of the shared calculations.
        parameters:
            Parameters for the calculation of the metric.
        n_jobs:
            Number of cpus for multi processing computations. Set to -1 to use all available cpus.
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
    """

    if metric not in get_metric_names():
        raise ValueError(f"Unsupported metric: {metric}")

    progress, started = init_progress(progress, verbosity, level)

    description = "Computing final compuations for " + metric + "..."

    if metric == "cluster_similarity":
        ann = pre_results
        ref_ann = shared_results
        nmis = clustering_nmis(
            ann, ref_ann, parameters["ns"], progress=progress, level=level, verbosity=verbosity, description=description
        )
        return nmis
    elif metric == "knn_overlap":
        knn_df = pre_results
        ref_knn_df = shared_results
        return mean_overlaps(
            knn_df,
            ref_knn_df,
            parameters["ks"],
            progress=progress,
            level=level,
            verbosity=verbosity,
            description=description,
        )
    elif metric == "forest_clfs":
        results = xgboost_forest_classification(
            adata,
            genes,
            celltypes="all",
            ct_key=parameters["ct_key"],
            n_jobs=n_jobs,
            progress=progress,
            level=level,
            verbosity=verbosity,
            description=description,
        )
        conf_mat = results[0]
        return conf_mat
    elif metric == "marker_corr":
        marker_cor = shared_results
        params = {k: v for k, v in parameters.items() if (k != "marker_list")}
        return max_marker_correlations(
            genes, marker_cor, **params, progress=progress, level=level, verbosity=verbosity, description=description
        )
    elif metric == "gene_corr":
        full_cor_mat = shared_results
        return gene_set_correlation_matrix(
            genes,
            full_cor_mat,
            ordered=True,
            progress=progress,
            level=level,
            verbosity=verbosity,
            description=description,
        )

    if progress and started:
        progress.stop()


def metric_summary(results: pd.DataFrame = None, metric: str = None, parameters: Dict = {}) -> Dict[str, Any]:
    """Simple final computations (per probe set) that aggregate evaluations to summary metrics.

    Args:
        results:
            The results of the previously calculated metric.
        metric:
            The name of the previously calculated metric.
        parameters:
            Parameters for the calculation of the metric.
    """

    if metric not in get_metric_names():
        raise ValueError(f"Unsupported metric: {metric}")

    summary: Dict[str, Any] = {}

    if metric == "cluster_similarity":
        nmis = results
        for s, val in summary_nmi_AUCs(nmis, parameters["AUC_borders"]).items():
            summary["cluster_similarity " + s] = val
    elif metric == "knn_overlap":
        means_df = results
        summary["knn_overlap mean_overlap_AUC"] = summary_knn_AUC(means_df)
    elif metric == "forest_clfs":
        conf_mat = results
        summary["forest_clfs accuracy"] = summary_metric_diagonal_confusion_mean(conf_mat)
        if "threshold" in parameters:
            th = parameters["threshold"]
            summary[f"forest_clfs perct acc > {th}"] = summary_metric_diagonal_confusion_percentage(
                conf_mat, threshold=th
            )
    elif metric == "marker_corr":
        cor_df = results
        for s, val in summary_marker_corr(cor_df).items():
            summary["marker_corr " + s] = val
    elif metric == "gene_corr":
        cor_mat = results
        summary["gene_corr 1 - mean"] = summary_metric_correlation_mean(cor_mat)
        if "threshold" in parameters:
            th = parameters["threshold"]
            summary[f"gene_corr perct max < {th}"] = summary_metric_correlation_percentage(cor_mat, threshold=th)

    return summary


#######################################
# cluster_similarity metric functions #
#######################################

# SHARED AND and PER PROBESET computations


def compute_clustering_and_update(
    adata: sc.AnnData, annotations: pd.DataFrame, resolution: float, tried_res_n: List[List], found_ns: List[int]
) -> Tuple[pd.DataFrame, List[List], List[int]]:
    """Compute a new leiden clustering and save the used resolution and resulting clusters.

    Args:
        adata:
            An already preprocessed annotated data matrix. Typically we use log normalised data.
        annotations:
            Contains clusterings for different numbers of clusters. Each row is a list of cluster assignments::

                n , obs1, obs2, obs3, ...
                2 ,   0 ,   0 ,   1 , ...
                4 ,   1 ,   0 ,   3 , ...
                8 ,   7 ,   7 ,   2 , ...
                3 ,   2 ,   1 ,   0 , ...

        resolution:
            Resolution to start computing clusterings.
        tried_res_n:
            Resolutions that were previously used for computing clusters.
        found_ns:
            The numbers of clusters resulting from previous clusters with the resolutions saved in :attr:`tried_res_n`.
    """
    sc.tl.leiden(adata, resolution=resolution, key_added="tmp")
    n = len(set(adata.obs["tmp"]))
    if n not in found_ns:
        annotations.loc[n, adata.obs_names] = list(adata.obs["tmp"])
        annotations.loc[n, "resolution"] = resolution
        found_ns.append(n)
    tried_res_n.append([resolution, n])
    tried_res_n.sort(key=lambda x: x[0])
    return annotations, tried_res_n, found_ns


def leiden_clusterings(
    adata: sc.AnnData,
    ns: Union[range, List[int]],
    start_res: float = 1.0,
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
    description: str = "Leiden clusterings...",
) -> pd.DataFrame:
    """Compute leiden clusters for different numbers of clusters.

    Leiden clusters are calculated with different resolutions.
    A search (similar to binary search) is applied to find the right resolutions for all defined n's.

    Args:
        adata:
            Adata object with data to compute clusters on.
        ns:
            The minimum (:attr:`ns[0]`) and maximum (:attr:`ns[1]`) number of clusters.
        start_res:
            Resolution to start computing clusterings.
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
        description:
            Description for progress bar.

    Returns:
        pd.DataFrame
            1st column refers to the number of clusters,
            2nd col refers to the resolution used to calculate the clustering
            each following column refer to individual cell's cluster assignments e.g.::

                n , res  , <adata.obs.index[0]>, <adata.obs.index[1]>, ...., <adata.obs.index[n_cells-1]>
                6 , 1.   ,         0           ,          4          , ....,              5
                2 , 0.67 ,         1           ,          0          , ....,              1
                13, 2.34 ,         9           ,          7          , ....,             12
                17, 2.78 ,         7           ,          7          , ....,              3
                .
                .
                .
    """

    progress, started = init_progress(progress, verbosity, level)
    if progress:
        task_leiden = progress.add_task(description, level=level, total=1)

    # Convert min and max n to list of ns
    if len(ns) != 2:
        raise ValueError("`ns` must be a list of two integers.")
    ns = range(ns[0], ns[1] + 1)

    # Clean adata and recalculate pca + neighbors graph
    a = adata.copy()
    clean_adata(a)
    sc.tl.pca(a)
    sc.pp.neighbors(a)

    # Initialize
    annotations = pd.DataFrame(index=ns, columns=["resolution"] + a.obs_names.tolist())
    annotations.index.name = "n"
    tried_res_n: List[List] = []
    found_ns: List[int] = []

    # First clustering step
    n_min = np.min(ns)
    n_max = np.max(ns)
    res = start_res
    annotations, tried_res_n, found_ns = compute_clustering_and_update(a, annotations, res, tried_res_n, found_ns)

    # Search for lower resolution border
    while np.min(np.unique([res_n[1] for res_n in tried_res_n])) > n_min:
        res *= 0.5
        annotations, tried_res_n, found_ns = compute_clustering_and_update(a, annotations, res, tried_res_n, found_ns)

    # Search for higher resolution border
    res = np.max([res_n[0] for res_n in tried_res_n])
    while np.max([res_n[1] for res_n in tried_res_n]) < n_max:
        res *= 2
        annotations, tried_res_n, found_ns = compute_clustering_and_update(a, annotations, res, tried_res_n, found_ns)

    # Search missing n's between neighboring found n's
    found_space = True
    while (not set(ns) <= set([res_n[1] for res_n in tried_res_n])) and (found_space):
        tmp_res_n = tried_res_n
        found_space = False
        for i in range(len(tmp_res_n) - 1):
            # check if two neighboring resolutions have different n's
            cond1 = tmp_res_n[i + 1][1] - tmp_res_n[i][1] > 1
            # check if we search an n between the two n's of the given resolutions
            cond2 = len([n for n in ns if n > tmp_res_n[i][1] and n < tmp_res_n[i + 1][1]]) > 0
            # we run in some error to recompute values between false pairs
            # I think this occurs since it can happen that sometimes a slightly higher
            # resolution leads to a lower number of clusters
            # TODO: check this properly
            # cond3 is added to account for that issue (probably not the best way)
            cond3 = abs(tmp_res_n[i + 1][0] - tmp_res_n[i][0]) > 0.00005  # 0.0003
            if cond1 and cond2 and cond3:
                res = (tmp_res_n[i][0] + tmp_res_n[i + 1][0]) * 0.5
                annotations, tried_res_n, found_ns = compute_clustering_and_update(
                    a, annotations, res, tried_res_n, found_ns
                )
                found_space = True

    if progress:
        progress.advance(task_leiden)
        if started:
            progress.stop()

    return annotations


def clustering_nmis(
    annotations: pd.DataFrame,
    ref_annotations: pd.DataFrame,
    ns: Union[range, List[int]],
    method: str = "arithmetic",
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
    description: str = "Cluster NMIs...",
) -> pd.DataFrame:
    """Compute NMI between clusterings and a reference set of clusterings.

    For different numbers of clusters (`ns`) the normalized mutual information
    NMI based on 2 different cluster annotations are computed.

    Args:
        annotations:
            Contains clusterings for different numbers of clusters. Each row is a list of cluster assignments::

                n , obs1, obs2, obs3, ...
                2 ,   0 ,   0 ,   1 , ...
                4 ,   1 ,   0 ,   3 , ...
                8 ,   7 ,   7 ,   2 , ...
                3 ,   2 ,   1 ,   0 , ...

        ref_annotations:
            Same as annotations for reference clusterings.
        ns:
            Minimum (:attr:`ns[0]`) and maximum (:attr:`ns[1]`) number of clusters.
        method:
            NMI implementation

                - 'max': scikit method with `average_method='max'`
                - 'min': scikit method with `average_method='min'`
                - 'geometric': scikit method with `average_method='geometric'`
                - 'arithmetic': scikit method with `average_method='arithmetic'`

            TODO: implement the following (see comment below and scib)

                - 'Lancichinetti': implementation by A. Lancichinetti 2009 et al.
                - 'ONMI': implementation by Aaron F. McDaid et al. (https://github.com/aaronmcdaid/Overlapping-NMI) Hurley 2011

        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
        description:
            Description for progress bar.

    Returns:
        pd.DataFrame of NMI results::

            n (index), nmi
            2        , 1.0
            3        , 0.9989
            ...
    """

    from sklearn.metrics import normalized_mutual_info_score

    progress, started = init_progress(progress, verbosity, level)

    # Convert min and max n to list of ns
    if len(ns) != 2:
        raise ValueError("`ns` must be a list of two integers.")
    ns = range(ns[0], ns[1] + 1)

    nmis = pd.DataFrame(np.nan, index=ns, columns=["nmi"])

    # Drop column "resolution"
    ann = annotations[[c for c in annotations.columns if (c != "resolution")]].copy()
    ref_ann = ref_annotations[[c for c in ref_annotations.columns if (c != "resolution")]].copy()

    # Prepare shared ns
    found_ns_df = ~ann.isnull().any(axis=1)
    ref_found_ns_df = ~ref_ann.isnull().any(axis=1)
    found_ns = found_ns_df.loc[found_ns_df].index.tolist()
    ref_ns = ref_found_ns_df.loc[ref_found_ns_df].index.tolist()
    valid_ns = [n for n in found_ns if (n in ref_ns) and (n in ns)]

    if progress:
        task_nmi = progress.add_task(description, total=len(valid_ns), level=level)

    # Calculate nmis
    for n in valid_ns:
        labels = ann.loc[n].values
        ref_labels = ref_ann.loc[n].values
        nmis.loc[n, "nmi"] = normalized_mutual_info_score(labels, ref_labels, average_method=method)

        if progress:
            progress.advance(task_nmi)

    if progress and started:
        progress.stop()

    return nmis


# SUMMARY metrics
def AUC(series: pd.Series, n_min: int = 1, n_max: int = 60) -> float:
    """Calculate the Area Unter the Curve.

    Args:
        series:
            Series of values.
        n_min:
            Lower border of curve.
        n_max:
            Upper border of curve.
    """
    tmp = series.loc[(series.index >= n_min) & (series.index <= n_max)]
    n = len(tmp)
    return tmp.sum() / n


def summary_nmi_AUCs(nmis: pd.DataFrame, AUC_borders: List[List]) -> Dict[str, float]:
    """Calculate AUC over range of nmi values.

    Args:
        nmis:
            Table of NMI results::

                n (index), nmi
                2        , 1.0
                3        , 0.9643
                4        , NaN
                5        , 0.98
                ...

        AUC_borders:
            Calculates nmi AUCs over given borders. E.g. :attr:`AUC_borders = [[2,4],[5,20]]` calculates nmi over n
            ranges 2 to 4 and 5 to 20. Defined border shouldn't exceed values in :attr:`nmis`.
    """
    AUCs = {}
    for ths in AUC_borders:
        AUCs[f"nmi_{ths[0]}_{ths[1]}"] = AUC(nmis["nmi"].interpolate(), n_min=ths[0], n_max=ths[1])
    return AUCs


################################
# knn_overlap metric functions #
################################

# SHARED AND and PER PROBESET computations


def knns(
    adata: sc.AnnData,
    genes: Union[List, str] = "all",
    ks: List[int] = [10, 20],
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
    description: str = "KNNS...",
) -> pd.DataFrame:
    """Compute nearest neighbors of observations for different ks.

    Args:
        adata:
            An already preprocessed annotated data matrix. Typically we use log normalised data.
        genes:
            A list of selected genes or "all".
        ks:
            Calculate knn graphs for each k in :attr:`ks`.
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
        description:
            Description for progress bar.

    Returns:
        pd.DataFrame:
            Includes nearest neighbors for all ks::

                gene (index), k10_1, k10_2, ..., k20_1, k20_2, ...
                ISG15       , 3789 , 512  ,    , 9720 , 15   , ...
                TNFRSF4     , 678  , 713  ,    , 7735 , 6225 , ...
                ...
    """

    progress, started = init_progress(progress, verbosity, level)

    # Subset adata to gene set
    if isinstance(genes, str) and (genes == "all"):
        genes = adata.var_names
    a = adata[:, genes].copy()

    # Set n_pcs to 50 or number of genes if < 50
    n_pcs = np.min([50, len(genes) - 1])

    # Delete existing PCAs, neighbor graphs, etc. and calculate PCA for n_pcs
    uns = [key for key in a.uns]
    for u in uns:
        del a.uns[u]
    obsm = [key for key in a.obsm]
    for o in obsm:
        del a.obsm[o]
    varm = [key for key in a.varm]
    for v in varm:
        del a.varm[v]
    obsp = [key for key in a.obsp]
    for o in obsp:
        del a.obsp[o]
    sc.tl.pca(a, n_comps=n_pcs)  # use_highly_variable=False

    # Get nearest neighbors for each k
    df = pd.DataFrame(index=a.obs_names)
    if progress:
        task_knn = progress.add_task(description, level=level, total=len(ks))
    for k in ks:
        if "neighbors" in a.uns:
            del a.uns["neighbors"]
        if "connectivities" in a.obsp:
            del a.obsp["connectivities"]
        if "distances" in a.obsp:
            del a.obsp["distances"]
        sc.pp.neighbors(a, n_neighbors=k)
        rows, cols = a.obsp["distances"].nonzero()
        nns = []
        for r in range(a.n_obs):
            nns.append(cols[rows == r].tolist())
        nn_df = pd.DataFrame(nns, index=a.obs_names)
        nn_df.columns = [f"k{k}_{i}" for i in range(len(nn_df.columns))]
        df = pd.concat([df, nn_df], axis=1)

        if progress:
            progress.advance(task_knn)

    if progress and started:
        progress.stop()

    return df


def mean_overlaps(
    knn_df: pd.DataFrame,
    ref_knn_df: pd.DataFrame,
    ks: List[int],
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
    description: str = "Mean overlaps...",
) -> pd.DataFrame:
    """Calculate mean overlaps of knn graphs of different ks.

    Args:
        knn_df:
            The results of the knn calculations, that are independent of the shared calculations.
        ref_knn_df:
            The results of the metric calculations, that are not probe set specific.
        ks:
            Calculate knn graphs for each k in `ks`.
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
        description:
            Description for progress bar.


    Returns:
        pd.DataFrame:
            k (index), mean

    """

    progress, started = init_progress(progress, verbosity, level)

    if progress:
        task_meano_verlap = progress.add_task(description, level=level, total=len(ks))
    df = pd.DataFrame(index=ks, data={"mean": 0.0})
    for k in ks:
        overlaps = []
        cols_of_k = [col for col in knn_df.columns if (int(col.split("_")[0][1:]) == k)]
        ref_cols_of_k = [col for col in ref_knn_df.columns if (int(col.split("_")[0][1:]) == k)]
        nns1 = knn_df[cols_of_k].values.astype(int)
        nns2 = ref_knn_df[ref_cols_of_k].values.astype(int)
        for i in range(nns1.shape[0]):
            set1 = set(nns1[i])
            set2 = set(nns2[i])
            max_intersection = np.min([len(set1), len(set2)])
            overlaps.append(len(set1.intersection(set2)) / max_intersection)
        df.loc[k, "mean"] = np.mean(overlaps)

        if progress:
            progress.advance(task_meano_verlap)

    if progress and started:
        progress.stop()

    return df


# SUMMARY metrics


def summary_knn_AUC(means_df: pd.DataFrame) -> float:
    """Calculate AUC of mean overlaps over ks.

    Args:
        means_df: The results of the previously calculated metric (knn_overlap mean_overlap_AUC).
    """
    x = [int(x) for x in means_df.index.values]
    y = means_df["mean"].values
    tmp = pd.Series(index=range(np.min(x), np.max(x)), dtype="float64")
    for i in range(len(y)):
        tmp.loc[x[i]] = y[i]
    return AUC(tmp.interpolate(), n_min=np.min(x), n_max=np.max(x))


################################
# forest_clfs metric functions #
################################

# SHARED computations
# None

# PER PROBESET computations
def xgboost_forest_classification(
    adata: sc.AnnData,
    selection: Union[List, pd.DataFrame],
    celltypes: Union[List, str] = "all",
    ct_key: str = "Celltypes",
    n_cells_min: int = 40,
    max_depth: int = 3,
    lr: float = 0.2,
    colsample_bytree: float = 1,
    cv_splits: int = 5,
    min_child_weight: Union[float, None] = None,
    gamma: float = None,
    seed: int = 0,
    n_seeds: int = 5,
    verbosity: int = 0,
    return_train_perform: bool = False,
    return_clfs: bool = False,
    return_predictions: bool = False,
    n_jobs: int = 1,
    progress: Progress = None,
    level: int = 3,
    description: str = "XGB forest classification",
) -> List:
    """Measure celltype classification performance with gradient boosted forests.

    We train extreme gradient boosted forest classifiers on multi class classification of cell types. Cross validation
    is performed to get an average confusion matrix (normalised by ground truth counts and sample weights to weight cell
    types in a balanced way). To make the performance measure robust only cell types with at least :attr:`n_cells_min`
    are taken into account. To make the cross validation more robust we run it with :attr:`n_seeds`. I.e.
    :attr:`cv_splits x :attr:n_seeds` classifiers are trained and evaluated.

    Args:
        adata:
            An already preprocessed annotated data matrix. Typically we use log normalised data.
        selection:
            Forests are trained on genes of the list or genes defined in the bool column selection['selection'].
        celltypes:
            Forests are trained on the given celltypes.
        ct_key:
            Column name of adata.obs with cell type info.
        n_cells_min:
            Minimal number of cells to filter out cell types from the training set. Performance results are not robust
            for low :attr:`n_cells_min`.
        max_depth:
            The max_depth argument of XGBClassifier.
        lr:
            Learning rate of XGBClassifier.
        colsample_bytree:
            Fraction of features (randomly selected) that will be used to train each tree of XGBClassifier.
        cv_splits:
            Number of cross validation splits.
        min_child_weight:
            Minimum sum of instance weight(hessian) needed in a child.
        gamma:
            Regularisation parameter of XGBClassifier. Instruct trees to add nodes only if the associated loss gain is
            larger or equal to gamma.
        seed:
            Random number seed.
        n_seeds:
            Number of training repetitions with different seeds. We use multiple seeds to make results more robust.
            Also we don't want to increase the number of CV splits to keep a minimal test size.
        verbosity:
            Set to 2 for progress bar. Set to 3 to print test performance of each tree during training.
        return_train_perform:
            Wether to also return confusion matrix of training set.
        return_clfs:
            Wether to return the classifier objects.
        return_predictions:
            Whether to return a list of prediction dataframes
        n_jobs:
            Multiprocessing number of processes.
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        description:
            Description of progress bar.


    Returns:
        pd.DataFrame:
            confusion matrix averaged over cross validations and seeds.
        pd.DataFrame:
            confusion matrix standard deviation over cross validations and seeds.
        if return_train_perform:
            pd.DataFrames as above for train set.
        if return_clfs:
            list of XGBClassifier objects of each cross validation step and seed.
        if return_predictions:
            list of dataframes with prediction results.
    """

    # if verbosity > 1:
    #     try:
    #         from tqdm.notebook import tqdm
    #     except ImportError:
    #         from tqdm import tqdm_notebook as tqdm
    #     desc = "XGBClassifier Cross Val."
    # else:
    #     tqdm = None

    progress, started = init_progress(progress, verbosity, level)

    # Define cell type list
    if celltypes == "all":
        celltypes = adata.obs[ct_key].unique().tolist()
    # Filter out cell types with less cells than n_cells_min
    cell_counts = adata.obs[ct_key].value_counts().loc[celltypes]
    if (cell_counts < n_cells_min).any() and (verbosity > 0):
        print(
            f"[bold yellow]The following cell types are not included in forest classifications since they have fewer "
            f"than {n_cells_min} cells: {cell_counts.loc[cell_counts < n_cells_min].index.tolist()}"
        )
        celltypes = [ct for ct in celltypes if (cell_counts.loc[ct] >= n_cells_min)]

    # Get data
    obs = adata.obs[ct_key].isin(celltypes)
    s = selection if isinstance(selection, list) else selection.loc[selection["selection"]].index.tolist()
    X = adata[obs, s].X
    ct_encoding = {ct: i for i, ct in enumerate(celltypes)}
    y = adata.obs.loc[obs, ct_key].astype(str).map(ct_encoding).values

    # Seeds
    # To have more robust results we use multiple seeds (especially important for cell types with low cell count)
    if n_seeds > 1:
        rng = np.random.default_rng(seed)
        seeds = list(rng.integers(low=0, high=100000, size=n_seeds))
    else:
        seeds = [seed]

    # Initialize variables for training
    confusion_matrices = []
    if return_train_perform:
        confusion_matrices_train = []
    if return_clfs:
        clfs = []
    if return_predictions:
        pred_dfs = []
        obs_names = np.array(adata.obs_names[obs])
        raw_df = pd.DataFrame(
            index=obs_names,
            data={
                "celltype": adata.obs.loc[obs, ct_key],
                "label": y,
                "train": False,
                "test": False,
                "pred": 0,
                "correct": False,
            },
        )
    n_classes = len(celltypes)

    # Cross validated random forest training
    # for seed in tqdm(seeds, desc="seeds", total=len(seeds)) if tqdm else seeds:
    if progress:
        task_forest = progress.add_task(description=description, total=len(seeds) * cv_splits, level=level)
    for seed in seeds:
        k_fold = StratifiedKFold(n_splits=cv_splits, random_state=seed, shuffle=True)
        # for train_ix, test_ix in tqdm(k_fold.split(X, y), desc=desc, total=cv_splits) if tqdm else k_fold.split(X, y):
        for train_ix, test_ix in k_fold.split(X, y):
            # Get train and test sets
            train_x, train_y, test_x, test_y = X[train_ix], y[train_ix], X[test_ix], y[test_ix]
            sample_weight_train = compute_sample_weight("balanced", train_y)
            sample_weight_test = compute_sample_weight("balanced", test_y)
            # Fit the classifier
            n_classes = len(np.unique(train_y))
            clf = XGBClassifier(
                max_depth=max_depth,
                num_class=n_classes if n_classes > 2 else None,
                n_estimators=250,
                objective="multi:softmax" if n_classes > 2 else "binary:logistic",
                early_stopping_rounds=5,
                eval_metric="mlogloss" if n_classes > 2 else "logloss",
                learning_rate=lr,
                colsample_bytree=colsample_bytree,
                min_child_weight=min_child_weight,
                gamma=gamma,
                booster="gbtree",  # TODO: compare with 'dart',rate_drop= 0.1
                random_state=seed,
                n_jobs=n_jobs,
            )
            clf.fit(
                train_x,
                train_y,
                sample_weight=sample_weight_train,
                eval_set=[(test_x, test_y)],
                sample_weight_eval_set=[sample_weight_test],
                verbose=verbosity > 2,
            )
            if return_clfs:
                clfs.append(clf)
            # Predict the labels of the test set samples
            y_pred = clf.predict(test_x)  # in case you try booster='dart' add, ntree_limit=1 (some value>0) check again
            # Append cv step results
            confusion_matrices.append(
                confusion_matrix(test_y, y_pred, normalize="true", sample_weight=sample_weight_test)
            )
            if return_predictions or return_train_perform:
                y_pred_train = clf.predict(train_x)
            if return_train_perform:
                confusion_matrices_train.append(
                    confusion_matrix(train_y, y_pred_train, normalize="true", sample_weight=sample_weight_train)
                )
            if return_predictions:
                df = raw_df.copy()
                df.loc[obs_names[train_ix], "train"] = True
                df.loc[obs_names[train_ix], "pred"] = y_pred_train
                df.loc[obs_names[test_ix], "test"] = True
                df.loc[obs_names[test_ix], "pred"] = y_pred
                df["correct"] = df["label"] == df["pred"]
                pred_dfs.append(df)

            if progress:
                progress.advance(task_forest)

    if started and progress:
        progress.stop()

    # Pool confusion matrices
    confusions_merged = np.concatenate([np.expand_dims(mat, axis=-1) for mat in confusion_matrices], axis=-1)
    confusion_mean = pd.DataFrame(index=celltypes, columns=celltypes, data=np.mean(confusions_merged, axis=-1))
    confusion_std = pd.DataFrame(index=celltypes, columns=celltypes, data=np.std(confusions_merged, axis=-1))
    if return_train_perform:
        confusions_merged = np.concatenate([np.expand_dims(mat, axis=-1) for mat in confusion_matrices_train], axis=-1)
        confusion_mean_train = pd.DataFrame(
            index=celltypes, columns=celltypes, data=np.mean(confusions_merged, axis=-1)
        )
        confusion_std_train = pd.DataFrame(index=celltypes, columns=celltypes, data=np.std(confusions_merged, axis=-1))

    # Return
    out = [confusion_mean, confusion_std]
    if return_train_perform:
        out += [confusion_mean_train, confusion_std_train]
    if return_clfs:
        out += [clfs]
    if return_predictions:
        out += [pred_dfs]
    return out


# SUMMARY metrics
def summary_metric_diagonal_confusion_mean(conf_mat: pd.DataFrame) -> np.ndarray:
    """Compute mean of diagonal elements of confusion matrix.

    Args:
        conf_mat:
            The results of the previously calculated metric (forest_clfs accuracy).
    """
    return np.diag(conf_mat).mean()


def linear_step(x: np.ndarray, low: float, high: float, descending: bool = True) -> np.ndarray:
    """Step function with linear transition between low and high.

    Args:
        x:
            Data.
        low:
            Lower border.
        high:
            Upper border.
        descending:
            Wether to go from 1 to 0 or the other way around.
    """

    b = 1.0
    m = 1 / (high - low)

    if descending:
        return np.where(x < low, b, np.where(x > high, 0, (x - low) * (-m) + b))
    else:
        return np.where(x < low, 0, np.where(x > high, b, (x - low) * m + 0))


def summary_metric_diagonal_confusion_percentage(
    conf_mat: pd.DataFrame, threshold: float = 0.9, tolerance: float = 0.05
) -> np.ndarray:
    """Compute percentage of diagonal elements of confusion matrix above threshold.

    Note:
        To make the metric more stable we smoothen the threshold with a linear transition from
        :attr:`threshold - tolerance` to :attr:`threshold + tolerance`.

    Args:
        conf_mat:
            The results of the previously calculated metric (forest_clfs accuracy).
        threshold:
            TODO add description
        tolerance:
            TODO add description
    """
    if tolerance:
        return np.mean(linear_step(np.diag(conf_mat), threshold - tolerance, threshold + tolerance, descending=False))
    else:
        return np.mean(np.diag(conf_mat) > threshold)


################################
# marker_corr metric functions #
################################

# SHARED computations
def marker_correlation_matrix(
    adata: sc.AnnData,
    marker_list: Union[str, Dict],
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
    description="Marker correlation...",
) -> pd.DataFrame:
    """Compute the correlation of each marker with all genes.

    Args:
        adata:
            An already preprocessed annotated data matrix. Typically we use log normalised data.
        marker_list:
            Either path to marker list or dict, e.g. {celltype1:[marker1,marker2], celltype2:[marker3,..], ..}.
            If a path is provided the marker list needs to be a csv formatted as::

                celltype1, celltype2, ...
                marker11,  marker21,
                marker12,  marker22,
                        ,  marker23,
                        ,  marker24,
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
        description:
            Description for progress bar.

    Returns:
        pd.DataFrame with genes of marker list as index

        columns:
            - "celltype" - cell type annotations of marker genes
            - "mean" - mean expression of markers
            - genes - for all genes of adata.var_names correlations with markers
    """

    full_cor_mat = correlation_matrix(
        adata, progress=progress, level=level, description=description, verbosity=verbosity
    )

    # Load marker_list as dict
    if isinstance(marker_list, str):
        marker_list_df = pd.read_csv(marker_list)
        marker_list_dict = dict_to_table(marker_list_df, genes_as_index=False, reverse=True)
    else:
        marker_list_dict = marker_list

    # Get markers and their corresponding celltype
    markers = [g for _, genes in marker_list_dict.items() for g in genes]
    markers = np.unique(markers).tolist()  # TODO: maybe throw a warning if genes occur twice
    # wrt celltype we take the first occuring celltype in marker_list for a given marker
    markers = [g for g in markers if g in adata.var_names]  # TODO: Throw warning if marker genes are not in adata
    ct_annot = []
    for g in markers:
        for ct, genes in marker_list_dict.items():
            if g in genes:
                ct_annot.append(ct)
                break  # to only add first celltype in list

    # Dataframes for calculations and potential min mean expression filters
    df = full_cor_mat.loc[markers]
    df_mean_tmp = gene_means(adata, genes=markers, key="mean", inplace=False)
    assert isinstance(df_mean_tmp, pd.DataFrame)
    df_mean = df_mean_tmp.loc[markers]
    # TODO: we might need to calculate the means at a previous point in case adata.X was scaled
    #      (or we just say this metric only makes sense on unscaled data if a min_mean is given)

    # Add mean expression and cell type annotation to dataframe
    df.insert(0, "mean", df_mean["mean"])
    ct_series = pd.Series(
        index=df.index, data=[ct for gene in df.index for ct, genes in marker_list_dict.items() if (gene in genes)]
    )
    df.insert(0, "celltype", ct_series)

    return df


# PER PROBESET computations
def max_marker_correlations(
    genes: List[str],
    marker_cor: pd.DataFrame,
    per_celltype: bool = True,
    per_marker: float = True,
    per_marker_min_mean: float = None,
    per_celltype_min_mean: float = None,
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
    description: str = "Max marker correlations...",
) -> pd.DataFrame:
    """Get maximal correlations with marker genes.

    Args:
        genes:
            The selected genes.
        marker_cor:
            Marker correlation matrix plus cell type annotations and mean expression (see output of
            :attr:`marker_correlation_matrix`).
        per_celltype:
            Wether to return columns with per cell type max correlations.
        per_marker:
            Wether to return columns with per marker max correlations.
        per_marker_min_mean:
            Add a column for correlation per marker that only takes into accounts markers with mean expression >
            min_mean_per_marker
        per_celltype_min_mean:
            Add a column for correlation per cell type that only takes into accounts markers with mean expression >
            min_mean_per_marker.
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
        description:
            Description for progress bar.

    Returns:
            pd.DataFrame with marker_genes as index

            columns:
                - "celltype":
                    cell type annotations
                - "mean":
                    mean expression of marker genes
                - max correlations:
                    - "per marker" maximal correlation of probeset and marker
                    - "per celltype" only highest correlation per cell type is not nan
                    - f"... mean > {min_mean}" filtered out markers with mean expression <= min_mean
    """

    progress, started = init_progress(progress, verbosity, level)

    if progress:
        task_max_corr = progress.add_task(description, level=level, total=1)

    cor_df = marker_cor[["celltype", "mean"]].copy()
    cor_df["per marker"] = marker_cor[genes].max(axis=1)
    if per_celltype:
        cor_df["per celltype"] = cor_df["per marker"]
        idxs = cor_df.groupby(["celltype"])["per celltype"].transform(max) == cor_df["per celltype"]
        cor_df.loc[~idxs, "per celltype"] = np.nan

    if (per_marker_min_mean is not None) and per_marker:
        min_mean = per_marker_min_mean
        cor_df[f"per marker mean > {min_mean}"] = cor_df["per marker"]
        cor_df.loc[cor_df["mean"] <= min_mean, f"per marker mean > {min_mean}"] = np.nan

    if (per_celltype_min_mean is not None) and per_celltype:
        min_mean = per_celltype_min_mean
        col = f"per celltype mean > {min_mean}"
        cor_df[col] = cor_df["per marker"]
        cor_df.loc[cor_df["mean"] <= min_mean, col] = np.nan
        idxs = cor_df.groupby(["celltype"])[col].transform(max) == cor_df[col]
        cor_df.loc[~idxs, col] = np.nan

    if not per_marker:
        del cor_df["per marker"]

    if progress:
        progress.advance(task_max_corr)
        if started:
            progress.stop()

    return cor_df


# SUMMARY metrics
def summary_marker_corr(cor_df: pd.DataFrame) -> Dict:
    """Means of maximal correlations with marker genes:

    Args:
        cor_df:
            Table with maximal correlations with marker genes. :attr:`cor_df` typically has multiple columns where
            different correlations are filtered out. See :attr:`max_marker_correlations` output for expected
            :attr:`cor_df`.
    """
    summaries = cor_df[[col for col in cor_df.columns if (col != "mean")]].mean(axis=0)
    return {summary: val for summary, val in summaries.items()}


##############################
# gene_corr metric functions #
##############################

# SHARED computations
def correlation_matrix(
    adata: sc.AnnData,
    var_names: List = None,
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
    description="Correlation matrix...",
) -> pd.DataFrame:
    """Compute correlation matrix of adata.X

    Args:
        adata:
            An already preprocessed annotated data matrix. Typically we use log normalised data.
        var_names:
            Calculate correlation on subset of variables.
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
        description:
            Description for progress bar.

    Returns:
        pd.DataFrame where index and columns are adata.var_names or var_names subset if defined. Values are correlations
        between genes.
    """
    progress, started = init_progress(progress, verbosity=verbosity, level=level)

    if progress:
        task_corr = progress.add_task(description, level=level, total=1)

    if var_names:
        a = adata[:, var_names]
    else:
        a = adata

    if issparse(a.X):
        cor_mat = pd.DataFrame(index=a.var.index, columns=a.var.index, data=np.corrcoef(a.X.toarray(), rowvar=False))
    else:
        cor_mat = pd.DataFrame(index=a.var.index, columns=a.var.index, data=np.corrcoef(a.X, rowvar=False))

    if progress:
        progress.advance(task_corr)
        if started:
            progress.stop()

    return cor_mat


# PER PROBESET computations
def gene_set_correlation_matrix(
    genes: List,
    full_cor_mat: pd.DataFrame,
    ordered: bool = True,
    progress: Progress = None,
    level: int = 2,
    verbosity: int = 2,
    description: str = "Gene set correlation matrix...",
) -> pd.DataFrame:
    """Return (ordered) correlation matrix of genes.

    Args:
        genes:
            The selected genes.
        full_cor_mat:
            Correlation matrix of genes that include at least `genes`.
        ordered:
            Wether to order the correlation matrix by a linkage clustering.
        progress:
            :attr:`rich.Progress` object if progress bars should be shown.
        level:
            Progress bar level.
        verbosity:
            Verbosity level.
        description:
            Description for progress bar.
    """
    progress, started = init_progress(progress, verbosity, level)

    if progress:
        task_corr = progress.add_task(description, total=1, level=level)

    cor_mat = full_cor_mat.loc[genes, genes].copy()
    if ordered:
        cor_mat = cluster_corr(cor_mat)

    if progress:
        progress.advance(task_corr)
        if started:
            progress.stop()

    return cor_mat


# SUMMARY metrics
def summary_metric_correlation_mean(cor_matrix: pd.DataFrame) -> float:
    """Calculate 1 - mean correlation.

    Args:
        cor_matrix:
            Gene set correlation matrix.

    Returns:
        float:
            1 - mean correlation
    """
    cor_mat = cor_matrix.copy()
    cor_mat = np.abs(cor_mat)
    np.fill_diagonal(cor_mat.values, 0)
    cor_mat.values[~(np.arange(cor_mat.shape[0])[:, None] > np.arange(cor_mat.shape[1]))] = np.nan
    return 1 - np.nanmean(cor_mat)


def summary_metric_correlation_percentage(
    cor_matrix: pd.DataFrame, threshold: float = 0.8, tolerance: float = 0.05
) -> np.ndarray:
    """Calculate percentage of genes with max(abs(correlations)) < threshold.

    To make the metric more stable we smoothen the threshold with a linear transition from
    threshold - tolerance to threshold + tolerance.

    Args:
        cor_matrix:
            Gene set correlation matrix.
        threshold:
            TODO add description.
        tolerance:
            TODO add description

    Returns:
        float:
            percentage of genes with max(abs(correlations)) < threshold
    """
    cor_mat = cor_matrix.copy()
    cor_mat = np.abs(cor_mat)
    np.fill_diagonal(cor_mat.values, 0)
    if tolerance:
        return np.mean(
            linear_step(np.max(cor_mat, axis=0).values, threshold - tolerance, threshold + tolerance, descending=True)
        )
    else:
        return np.mean(np.max(cor_mat, axis=0).values < threshold)



#!/usr/bin/env python
"""Command-line interface."""
import logging
import sys

import click
import rich.logging
import spapros
from pypi_latest import PypiLatest
from rich import print
from rich import traceback
from spapros.evaluation.evaluation_pipeline import run_evaluation
from spapros.selection.selection import run_selection

log = logging.getLogger()
spapros_pypi_latest = PypiLatest("spapros", spapros.__version__)


def main() -> None:
    traceback.install()

    """Main entry point for spapros."""
    print(
        """[bold blue]
███████ ██████   █████  ██████  ██████   ██████  ███████ 
██      ██   ██ ██   ██ ██   ██ ██   ██ ██    ██ ██      
███████ ██████  ███████ ██████  ██████  ██    ██ ███████ 
     ██ ██      ██   ██ ██      ██   ██ ██    ██      ██ 
███████ ██      ██   ██ ██      ██   ██  ██████  ███████ 
                                                         
"""
    )
    print("[bold blue]Run [green]spapros --help [blue]for an overview of all commands\n")

    # Is the latest spapros version installed? Upgrade if not!
    if not spapros_pypi_latest.check_latest():
        print("[bold blue]Run [green]spapros upgrade [blue]to get the latest version.")
    spapros_cli(prog_name="spapros")


@click.group()
@click.version_option(
    spapros.__version__,
    message=click.style(f"spapros Version: {spapros.__version__}", fg="blue"),
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Enable verbose output (print debug statements).",
)
@click.option("-l", "--log-file", help="Save a verbose log to a file.")
@click.pass_context
def spapros_cli(ctx, verbose, log_file):
    """
    Create state of the art probesets for spatial transcriptomics projects.
    """
    # Set the base logger to output DEBUG
    log.setLevel(logging.DEBUG)

    # Set up logs to the console
    log.addHandler(
        rich.logging.RichHandler(
            level=logging.DEBUG if verbose else logging.INFO,
            console=rich.console.Console(file=sys.stderr),
            show_time=True,
            markup=True,
        )
    )

    # Set up logs to a file if we asked for one
    if log_file:
        log_fh = logging.FileHandler(log_file, encoding="utf-8")
        log_fh.setLevel(logging.DEBUG)
        log_fh.setFormatter(logging.Formatter("[%(asctime)s] %(name)-20s [%(levelname)-7s]  %(message)s"))
        log.addHandler(log_fh)


@spapros_cli.command()
@click.argument("data", type=click.Path(exists=True))
@click.option("--output", "-o", default="./results/")
def selection(data, output) -> None:
    """Create a selection of probesets for an h5ad file.

    Args:
        data: Path to the h5ad file
        output: Output path
    """
    run_selection(data, output)


@spapros_cli.command()
@click.argument("data", type=click.Path(exists=True))
@click.argument("probeset", type=click.Path(exists=True))
@click.argument("marker-file", type=click.Path(exists=True))
@click.argument("probeset-ids", nargs=-1)
@click.option("--parameters", type=click.Path(exists=True))
@click.option("--output", "-o", default="./results/")
def evaluation(data, probeset, marker_file, probeset_ids, parameters, output) -> None:
    """Create a selection of probesets for an h5ad file.

    Args:
        data: Path to the h5ad dataset file
        probeset: Path to the probeset file
        marker_file: Path to the marker file
        probeset_ids: Several probeset ids
        parameters: Path to a yaml file containing parameters
        output: Output path
    """
    if not probeset_ids:
        probeset_ids = "all"
    else:
        probeset_ids = list(probeset_ids)

    run_evaluation(
        adata_path=data,
        probeset=probeset,
        marker_file=marker_file,
        probeset_ids=probeset_ids,
        result_dir=output,
        parameters_file=parameters,
    )


@spapros_cli.command(short_help="Check for a newer version of ehrapy and upgrade if required.")
def upgrade() -> None:
    """Checks whether the locally installed version of spapros is the latest & upgrades if not."""
    spapros_pypi_latest.check_upgrade()


if __name__ == "__main__":
    main()  # pragma: no cover



"""FRom squidpy's general utility functions."""
from __future__ import annotations

from enum import Enum
from multiprocessing import cpu_count
from multiprocessing import Manager
from queue import Queue
from threading import Thread
from typing import Any
from typing import Callable
from typing import Hashable
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import joblib as jl
import numpy as np

# from contextlib import contextmanager
# from typing import Generator

__all__ = ["singledispatchmethod", "Signal", "SigQueue"]


try:
    from functools import singledispatchmethod
except ImportError:
    from functools import singledispatch, update_wrapper

    def singledispatchmethod(func: Callable[..., Any]) -> Callable[..., Any]:  # type: ignore[no-redef]
        """Backport of `singledispatchmethod` for < Python 3.8."""
        dispatcher = singledispatch(func)

        def wrapper(*args: Any, **kw: Any) -> Any:
            return dispatcher.dispatch(args[1].__class__)(*args, **kw)

        wrapper.register = dispatcher.register  # type: ignore[attr-defined]
        update_wrapper(wrapper, func)

        return wrapper


class SigQueue(Queue["Signal"] if TYPE_CHECKING else Queue):  # type: ignore[misc]
    """Signalling queue."""


def _unique_order_preserving(iterable: Iterable[Hashable]) -> Tuple[List[Hashable], Set[Hashable]]:
    """Remove items from an iterable while preserving the order."""
    seen: Set[Hashable] = set()
    seen_add = seen.add
    return [i for i in iterable if not (i in seen or seen_add(i))], seen


class Signal(Enum):
    """Signalling values when informing parallelizer."""

    NONE = 0
    UPDATE = 1
    FINISH = 2
    UPDATE_FINISH = 3


def parallelize(
    callback: Callable[..., Any],
    collection: Sequence[Any],
    n_jobs: int = 1,
    n_split: Optional[int] = None,
    unit: str = "",
    desc: Optional[str] = None,
    use_ixs: bool = False,
    backend: str = "loky",
    extractor: Optional[Callable[[Sequence[Any]], Any]] = None,
    show_progress_bar: bool = True,
    use_runner: bool = False,
    **_: Any,
) -> Any:
    """
    Parallelize function call over a collection of elements.
    Parameters
    ----------
    callback
        Function to parallelize. Can either accept a whole chunk (``use_runner=False``) or just a single
        element (``use_runner=True``).
    collection
        Sequence of items to split into chunks.
    n_jobs
        Number of parallel jobs.
    n_split
        Split ``collection`` into ``n_split`` chunks.
        If <= 0, ``collection`` is assumed to be already split into chunks.
    unit
        Unit of the progress bar.
    desc
        Description of the progress bar.
    use_ixs
        Whether to pass indices to the callback.
    backend
        Which backend to use for multiprocessing. See class `joblib.Parallel` for valid options.
    extractor
        Function to apply to the result after all jobs have finished.
    show_progress_bar
        Whether to show a progress bar.
    use_runner
        Whether the ``callback`` handles only 1 item from the ``collection`` or a chunk.
        The latter grants more control, e.g. using func `numba.prange` instead of normal iteration.
    Returns
    -------
    The result depending on ``callable``, ``extractor``.
    """
    if show_progress_bar:
        try:
            from tqdm.notebook import tqdm
        except ImportError:
            from tqdm import tqdm_notebook as tqdm
    else:
        tqdm = None

    def runner(iterable: Iterable[Any], *args: Any, queue: Optional["SigQueue"] = None, **kwargs: Any) -> List[Any]:
        result: List[Any] = []

        for it in iterable:
            res = callback(it, *args, **kwargs)
            if res is not None:
                result.append(result)
            if queue is not None:
                queue.put(Signal.UPDATE)

        if queue is not None:
            queue.put(Signal.FINISH)

        return result

    def update(pbar: "tqdm.std.tqdm", queue: "SigQueue", n_total: int) -> None:
        n_finished = 0
        while n_finished < n_total:
            try:
                res = queue.get()
            except EOFError as e:
                if not n_finished != n_total:
                    raise RuntimeError(f"Finished only `{n_finished}` out of `{n_total}` tasks.") from e
                break

            # assert isinstance(res, Signal), f"Invalid type `{type(res).__name__}`."

            if res in (Signal.FINISH, Signal.UPDATE_FINISH):
                n_finished += 1
            if pbar is not None and res in (Signal.UPDATE, Signal.UPDATE_FINISH):
                pbar.update()

        if pbar is not None:
            pbar.close()

    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if pass_queue and show_progress_bar:
            pbar = None if tqdm is None else tqdm(total=col_len, unit=unit, desc=desc)
            queue = Manager().Queue()
            thread = Thread(target=update, args=(pbar, queue, len(collections)))
            thread.start()
        else:
            pbar, queue, thread = None, None, None  # type: ignore[assignment]

        res = jl.Parallel(n_jobs=n_jobs, backend=backend)(
            jl.delayed(runner if use_runner else callback)(
                *((i, cs) if use_ixs else (cs,)),
                *args,
                **kwargs,
                queue=queue,
            )
            for i, cs in enumerate(collections)
        )

        if thread is not None:
            thread.join()

        return res if extractor is None else extractor(res)

    if n_jobs == 0:
        raise ValueError("Number of jobs cannot be `0`.")
    if n_jobs < 0:
        n_jobs = cpu_count() + 1 + n_jobs

    if n_split is None:
        n_split = n_jobs

    if n_split <= 0:
        col_len = sum(map(len, collection))
        collections = collection
    else:
        col_len = len(collection)
        step = int(np.ceil(len(collection) / n_split))
        collections = list(filter(len, (collection[i * step : (i + 1) * step] for i in range(col_len))))

    if use_runner:
        use_ixs = False
    pass_queue = not hasattr(callback, "py_func")  # we'd be inside a numba function

    return wrapper


def _get_n_cores(n_cores: Optional[int]) -> int:
    """
    Make number of cores a positive integer.
    This is useful for especially logging.
    Parameters
    ----------
    n_cores
        Number of cores to use.
    Returns
    -------
    int
        Positive integer corresponding to how many cores to use.
    """
    if n_cores == 0:
        raise ValueError("Number of cores cannot be `0`.")
    if n_cores is None:
        return 1
    if n_cores < 0:
        return cpu_count() + 1 + n_cores

    return n_cores



from spapros.util.util import get_expression_quantile
from spapros.util.util import get_processed_pbmc_data
from spapros.util.util import plateau_penalty_kernel
from spapros.util.util import transfered_expression_thresholds

__all__ = [
    "get_expression_quantile",
    "transfered_expression_thresholds",
    "plateau_penalty_kernel",
    "get_processed_pbmc_data",
]



from spapros.selection.selection_methods import select_DE_genes
from spapros.selection.selection_methods import select_pca_genes
from spapros.selection.selection_procedure import ProbesetSelector
from spapros.selection.selection_procedure import select_reference_probesets

__all__ = [
    "ProbesetSelector",
    "select_reference_probesets",
    "select_pca_genes",
    "select_DE_genes",
]



from spapros.evaluation.evaluation import forest_classifications
from spapros.evaluation.evaluation import ProbesetEvaluator
from spapros.evaluation.evaluation import single_forest_classifications
from spapros.evaluation.metrics import get_metric_default_parameters

__all__ = [
    "get_metric_default_parameters",
    "forest_classifications",
    "single_forest_classifications",
    "ProbesetEvaluator",
]
